import abc
import datetime
from dataclasses import dataclass
from types import MappingProxyType
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from tecton._internals.tecton_pydantic import StrictModel
from tecton._internals.tecton_pydantic import pydantic_v1
from tecton._internals.type_utils import sdk_type_from_tecton_type
from tecton.aggregation_functions import AggregationFunction
from tecton.framework.configs import LifetimeWindow
from tecton.framework.configs import TimeWindow
from tecton.framework.configs import TimeWindowSeries
from tecton.framework.configs import build_aggregation_proto
from tecton.framework.expression_parser import expression_to_proto
from tecton.types import Field
from tecton.types import SdkDataType
from tecton_core.specs.feature_spec import FeatureMetadataSpec
from tecton_proto.args import feature_view__client_pb2 as feature_view_pb2
from tecton_proto.common import schema__client_pb2 as schema_pb2


class Feature(StrictModel, abc.ABC):
    description: Optional[str]
    tags: Optional[Dict[str, str]]

    @abc.abstractmethod
    def _to_proto(self, **kwargs):
        pass


class Aggregate(Feature):
    """
    The `Aggregate` class describes an aggregation feature that is applied to a Batch or Stream Feature View via `features` param.

    ```python
    from tecton import Aggregate, batch_feature_view, TimeWindow
    from tecton.types import Int64
    from datetime import timedelta

    @batch_feature_view(
        # ...
        features=[
            Aggregate(
                input_column=Field("my_column", Int64),
                function="mean",
                time_window=TimeWindow(window_size=timedelta(days=7)),
            ),
            Aggregate(
                input_column=Field("another_column", Int64),
                function="mean",
                time_window=TimeWindow(window_size=timedelta(days=1)),
                name="1d_average",
                description="my aggregate feature description",
                tags={"tag": "value"}
            ),
        ],
    )
    def my_fv(data_source):
        pass
    ```

    :param function: One of the built-in aggregation functions, such as "sum", "count", `last(2)` etc.
    :param time_window: The window_size and optional offset over which to aggregate over.
    :param name: The name of this feature. Defaults to an autogenerated name, e.g. transaction_count_7d_1d.
    :param input_column: Describes name and type of the column that will be used in the aggregation.
    :param description: A human-readable description of the feature
    :param tags: Tags associated with the feature (key-value pairs of user-defined metadata).
    """

    function: AggregationFunction
    time_window: Union[TimeWindow, TimeWindowSeries, LifetimeWindow]
    name: Optional[str] = None
    input_column: Field

    def __init__(
        self,
        function: Union[str, AggregationFunction],
        time_window: Union[TimeWindow, TimeWindowSeries, LifetimeWindow, datetime.timedelta],
        input_column: Field,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            function=function,
            time_window=time_window,
            name=name,
            input_column=input_column,
            description=description,
            tags=tags,
        )

    @pydantic_v1.validator("time_window", pre=True)
    def timedelta_to_time_window(cls, v):
        if isinstance(v, datetime.timedelta):
            return TimeWindow(window_size=v)
        return v

    @pydantic_v1.validator("function", pre=True)
    def str_to_aggregation_function(cls, v):
        if isinstance(v, str):
            return AggregationFunction(base_name=v, resolved_name=v, params=MappingProxyType({}))
        return v

    def _to_proto(
        self,
        aggregation_interval: datetime.timedelta,
        is_continuous: bool,
        compaction_enabled: bool = False,
        is_streaming_fv: bool = False,
    ) -> feature_view_pb2.FeatureAggregation:
        proto = build_aggregation_proto(
            self.name,
            self.input_column.name,
            self.function,
            self.time_window,
            aggregation_interval,
            is_continuous,
            compaction_enabled,
            is_streaming_fv,
            self.description,
            self.tags,
        )
        proto.column_dtype.CopyFrom(self.input_column.dtype.tecton_type.proto)
        return proto


class Attribute(Feature):
    """
    The `Attribute` class describes an attribute feature that is applied to a Feature View or Feature Table via `features` param.

    ```python
    from tecton import Attribute, batch_feature_view
    from tecton.types import String

    @batch_feature_view(
        # ...
        features=[
            Attribute(
                name="my_column",
                dtype=String
            )
            Attribute(
                name="my_other_column",
                dtype=String,
                description="my attribute feature description",
                tags={"tag": "value"}
            )
        ],
    )
    def my_fv(data_source):
        pass
    ```

    :param name: Name of a column from the transformation output.
    :param dtype: Datatype of a column from the transformation output
    :param description: A human-readable description of the feature
    :param tags: Tags associated with the feature (key-value pairs of user-defined metadata).
    """

    name: str
    dtype: SdkDataType

    def __init__(
        self, name: str, dtype: SdkDataType, description: Optional[str] = None, tags: Optional[Dict[str, str]] = None
    ):
        super().__init__(name=name, dtype=dtype, description=description, tags=tags)

    def _to_proto(self):
        return feature_view_pb2.Attribute(
            name=self.name, column_dtype=self.dtype.tecton_type.proto, description=self.description, tags=self.tags
        )


class Calculation(Feature):
    """
    The `Calculation` class describes a Calculation feature that is applied to a Realtime Feature View via the `features` param.

    ```python
    from tecton import Calculation, RealtimeFeatureView

    transaction_counts_fv = RealtimeFeatureView(
        name="transaction_counts_features",
        sources=[request_ds, user_transaction_metrics],
        # mode=None implied
        ...
        features = [
            Calculation(
                name="transaction_count_fill_na",
                expr="COALESCE(user_transaction_metrics.transaction_count, request_ds.default_count)",
            ),
            Calculation(
                name="transaction_count_add_1",
                expr="user_transaction_metrics.transaction_count + 1",
                description="a test calculation utilizing addition",
                tags={"tag": "value"}
            ),
        ],
    )
    ```

    :param name: The name of this feature. Must be explicitly defined.
    :param expr: The calculation string expressing the operations and input operands.
    :param description: A human-readable description of the feature.
    :param tags: Tags associated with the feature (key-value pairs of user-defined metadata).
    """

    name: str
    expr: str
    description: Optional[str] = None
    tags: Optional[Dict[str, str]] = None

    def __init__(
        self,
        name: str,
        expr: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        super().__init__(name=name, expr=expr, description=description, tags=tags)

    def _to_proto(self, input_schema: Dict[str, SdkDataType]):
        root = expression_to_proto(self.expr, input_schema)
        return feature_view_pb2.Calculation(
            name=self.name, expr=self.expr, abstract_syntax_tree_root=root, column_dtype=root.dtype
        )


class Embedding(Feature):
    """
    The `Embedding` class describes an embedding feature that is applied to a Batch Feature View via `features` param.


    ```python
    from tecton import Embedding, batch_feature_view
    from tecton.types import String

    @batch_feature_view(
        # ...
        features=[
            Embedding(
                input_column=Field("my_column", String),
                model="sentence-transformers/all-MiniLM-L6-v2",
                name="my_embedding_feature"
            ),
            Embedding(
                input_column=Field("my_other_column", String),
                model="sentence-transformers/all-MiniLM-L6-v2",
                name="my_other_embedding_feature",
                description="my embedding feature description",
                tags={"tag": "value"}
            )
        ],
    )
    def my_fv(data_source):
        return data_source
    ```
    :param input_column: Describes name and type of the column that will be used in the aggregation.
    :param model: The model name that is used to compute the embedding feature. Check https://docs.tecton.ai/docs/beta/defining-features/feature-views/embeddings for all supported models.
    :param name: The name of this feature. Defaults to an autogenerated name, e.g. "my_embedding_feature".
    :param description: A human-readable description of the feature
    :param tags: Tags associated with the feature (key-value pairs of user-defined metadata).
    """

    input_column: Field
    model: str
    name: Optional[str] = None

    def __init__(
        self,
        input_column: Field,
        model: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        super().__init__(model=model, name=name, input_column=input_column, description=description, tags=tags)

    def _to_proto(self) -> feature_view_pb2.Embedding:
        name = self.name or f"{self.input_column.name}_embedding"
        return feature_view_pb2.Embedding(
            name=name,
            column=self.input_column.name,
            column_dtype=self.input_column.dtype.tecton_type.proto,
            model=self.model,
            description=self.description,
            tags=self.tags,
        )


class Inference(Feature):
    """
    The `Inference` class describes a model inference feature that is applied to a Batch Feature View via `features` param.

    ```python
    from tecton import Inference, batch_feature_view
    from tecton.types import String, Int64

    @batch_feature_view(
        # ...
        features=[
            Inference(
                input_columns=[Field("my_field1", String), Field("my_field2", Int64)],
                model="my_custom_model",
                name="my_inference_feature"
            ),
            Inference(
                input_columns=[Field("my_field3", String), Field("my_field4", Int64)],
                model="my_custom_model",
                name="my_other_inference_feature"m
                description="my inference feature description",
                tags={"tag": "value"}
            )
        ],
    )
    def my_fv(data_source):
        return data_source
    ```

    :param input_columns: A list of columns that are used as inputs to the model.
    :param model: The model name that is used to compute the inference feature. The model needs to be registered in advance using `tecton model` CLI.
    :param name: The name of this feature. Defaults to an autogenerated name, e.g. "my_inference_feature".
    :param description: A human-readable description of the feature
    :param tags: Tags associated with the feature (key-value pairs of user-defined metadata).
    """

    model: str
    input_columns: List[Field]
    name: Optional[str]

    def __init__(
        self,
        model: str,
        input_columns: List[Field],
        name: str,
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        super().__init__(model=model, input_columns=input_columns, name=name, description=description, tags=tags)

    def _to_proto(self) -> feature_view_pb2.Inference:
        # TODO(jiadong): Figure out a default name for `Inference` feature.
        return feature_view_pb2.Inference(
            model=self.model,
            input_columns=[
                schema_pb2.Field(name=field.name, dtype=field.dtype.tecton_type.proto) for field in self.input_columns
            ],
            name=self.name,
            description=self.description,
            tags=self.tags,
        )


@dataclass
class FeatureMetadata:
    """
    `FeatureMetadata` contains metadata about your Feature.

    :param name: Name of the feature
    :param dtype: Data type of the feature
    :param description: A human-readable description of the feature
    :param tags: Tags associated with the feature (key-value pairs of user-defined metadata).
    """

    name: str
    dtype: SdkDataType
    description: Optional[str]
    tags: Optional[Dict[str, str]]

    @classmethod
    def _from_spec(cls, spec: FeatureMetadataSpec):
        return cls(
            name=spec.name, dtype=sdk_type_from_tecton_type(spec.dtype), description=spec.description, tags=spec.tags
        )



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class AutoscalingConfig(_message.Message):
    __slots__ = ["max_nodes", "min_nodes"]
    MAX_NODES_FIELD_NUMBER: _ClassVar[int]
    MIN_NODES_FIELD_NUMBER: _ClassVar[int]
    max_nodes: int
    min_nodes: int
    def __init__(self, min_nodes: _Optional[int] = ..., max_nodes: _Optional[int] = ...) -> None: ...

class ProvisionedScalingConfig(_message.Message):
    __slots__ = ["desired_nodes"]
    DESIRED_NODES_FIELD_NUMBER: _ClassVar[int]
    desired_nodes: int
    def __init__(self, desired_nodes: _Optional[int] = ...) -> None: ...



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.auditlog import metadata__client_pb2 as _metadata__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ComputeIdentity(_message.Message):
    __slots__ = ["databricks_service_principal"]
    DATABRICKS_SERVICE_PRINCIPAL_FIELD_NUMBER: _ClassVar[int]
    databricks_service_principal: DatabricksServicePrincipal
    def __init__(self, databricks_service_principal: _Optional[_Union[DatabricksServicePrincipal, _Mapping]] = ...) -> None: ...

class DatabricksServicePrincipal(_message.Message):
    __slots__ = ["application_id"]
    APPLICATION_ID_FIELD_NUMBER: _ClassVar[int]
    application_id: str
    def __init__(self, application_id: _Optional[str] = ...) -> None: ...

class Workspace(_message.Message):
    __slots__ = ["capabilities", "compute_identities", "created_at", "name"]
    CAPABILITIES_FIELD_NUMBER: _ClassVar[int]
    COMPUTE_IDENTITIES_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    capabilities: WorkspaceCapabilities
    compute_identities: _containers.RepeatedCompositeFieldContainer[ComputeIdentity]
    created_at: _timestamp_pb2.Timestamp
    name: str
    def __init__(self, name: _Optional[str] = ..., capabilities: _Optional[_Union[WorkspaceCapabilities, _Mapping]] = ..., compute_identities: _Optional[_Iterable[_Union[ComputeIdentity, _Mapping]]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class WorkspaceCapabilities(_message.Message):
    __slots__ = ["materializable", "offline_store_subdirectory_enabled"]
    MATERIALIZABLE_FIELD_NUMBER: _ClassVar[int]
    OFFLINE_STORE_SUBDIRECTORY_ENABLED_FIELD_NUMBER: _ClassVar[int]
    materializable: bool
    offline_store_subdirectory_enabled: bool
    def __init__(self, materializable: bool = ..., offline_store_subdirectory_enabled: bool = ...) -> None: ...



from tecton_proto.common import data_source_type__client_pb2 as _data_source_type__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import schema_container__client_pb2 as _schema_container__client_pb2
from tecton_proto.data import batch_data_source__client_pb2 as _batch_data_source__client_pb2
from tecton_proto.data import fco_metadata__client_pb2 as _fco_metadata__client_pb2
from tecton_proto.data import stream_data_source__client_pb2 as _stream_data_source__client_pb2
from tecton_proto.validation import validator__client_pb2 as _validator__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class VirtualDataSource(_message.Message):
    __slots__ = ["batch_data_source", "data_source_type", "fco_metadata", "options", "schema", "stream_data_source", "validation_args", "virtual_data_source_id"]
    class OptionsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    BATCH_DATA_SOURCE_FIELD_NUMBER: _ClassVar[int]
    DATA_SOURCE_TYPE_FIELD_NUMBER: _ClassVar[int]
    FCO_METADATA_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    STREAM_DATA_SOURCE_FIELD_NUMBER: _ClassVar[int]
    VALIDATION_ARGS_FIELD_NUMBER: _ClassVar[int]
    VIRTUAL_DATA_SOURCE_ID_FIELD_NUMBER: _ClassVar[int]
    batch_data_source: _batch_data_source__client_pb2.BatchDataSource
    data_source_type: _data_source_type__client_pb2.DataSourceType
    fco_metadata: _fco_metadata__client_pb2.FcoMetadata
    options: _containers.ScalarMap[str, str]
    schema: _schema_container__client_pb2.SchemaContainer
    stream_data_source: _stream_data_source__client_pb2.StreamDataSource
    validation_args: _validator__client_pb2.VirtualDataSourceValidationArgs
    virtual_data_source_id: _id__client_pb2.Id
    def __init__(self, virtual_data_source_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., batch_data_source: _Optional[_Union[_batch_data_source__client_pb2.BatchDataSource, _Mapping]] = ..., stream_data_source: _Optional[_Union[_stream_data_source__client_pb2.StreamDataSource, _Mapping]] = ..., data_source_type: _Optional[_Union[_data_source_type__client_pb2.DataSourceType, str]] = ..., fco_metadata: _Optional[_Union[_fco_metadata__client_pb2.FcoMetadata, _Mapping]] = ..., schema: _Optional[_Union[_schema_container__client_pb2.SchemaContainer, _Mapping]] = ..., validation_args: _Optional[_Union[_validator__client_pb2.VirtualDataSourceValidationArgs, _Mapping]] = ..., options: _Optional[_Mapping[str, str]] = ...) -> None: ...



import os
import shutil
import uuid
from unittest import TestCase

import pandas
import pyarrow
from deltalake.writer import write_deltalake

import tecton_core.tecton_pendulum as pendulum
from tecton_core import conf
from tecton_core.data_processing_utils import _calculate_split_count
from tecton_core.data_processing_utils import _even_split_pyarrow
from tecton_core.data_processing_utils import _get_total_row_number
from tecton_core.data_processing_utils import split_spine


pandas_df = pandas.DataFrame({"join_key": ["abc", "abd", "dbc", "dbe", "ace", "xyz", "xya", "xzz"]})
spine_pyarrow_table = pyarrow.table({"join_key": ["abc", "abd", "dbc", "dbe", "ace", "xyz", "xya", "xzz"]})
join_keys = ["join_key"]


class SpineSplitTest(TestCase):
    def test_spine_split_even(self):
        conf.set("DUCKDB_SPINE_SPLIT_STRATEGY", "even")
        conf.set("DUCKDB_SPINE_SPLIT_COUNT", 3)

        split_dfs = split_spine(pandas_df, join_keys)

        assert pandas.DataFrame.equals(split_dfs[0], pandas.DataFrame({"join_key": ["abc", "abd", "ace"]}))
        assert pandas.DataFrame.equals(split_dfs[1], pandas.DataFrame({"join_key": ["dbc", "dbe", "xya"]}))
        assert pandas.DataFrame.equals(split_dfs[2], pandas.DataFrame({"join_key": ["xyz", "xzz"]}))

    def test_spine_split_minimize_distance(self):
        conf.set("DUCKDB_SPINE_SPLIT_STRATEGY", "minimize_distance")
        conf.set("DUCKDB_SPINE_SPLIT_COUNT", 3)

        split_dfs = split_spine(pandas_df, join_keys)

        assert pandas.DataFrame.equals(split_dfs[0], pandas.DataFrame({"join_key": ["abc", "abd", "ace"]}))
        assert pandas.DataFrame.equals(split_dfs[1], pandas.DataFrame({"join_key": ["dbc", "dbe"]}))
        assert pandas.DataFrame.equals(split_dfs[2], pandas.DataFrame({"join_key": ["xya", "xyz", "xzz"]}))

    def test_spine_split_agglomerative_clustering(self):
        conf.set("DUCKDB_SPINE_SPLIT_STRATEGY", "agglomerative_clustering")
        conf.set("DUCKDB_SPINE_SPLIT_COUNT", 3)

        split_dfs = split_spine(pandas_df, join_keys)

        assert pandas.DataFrame.equals(split_dfs[0], pandas.DataFrame({"join_key": ["xyz", "xya", "xzz"]}))
        assert pandas.DataFrame.equals(split_dfs[1], pandas.DataFrame({"join_key": ["abc", "abd", "ace"]}))
        assert pandas.DataFrame.equals(split_dfs[2], pandas.DataFrame({"join_key": ["dbc", "dbe"]}))

    def test_spine_split_even_pyarrow(self):
        split_tables = _even_split_pyarrow(spine_pyarrow_table, join_keys, split_count=3)

        assert split_tables[0].equals(pyarrow.table({"join_key": ["abc", "abd", "ace"]}))
        assert split_tables[1].equals(pyarrow.table({"join_key": ["dbc", "dbe", "xya"]}))
        assert split_tables[2].equals(pyarrow.table({"join_key": ["xyz", "xzz"]}))

    def test_get_total_row_number(self):
        delta_table_path = f"/tmp/test_spine_split/delta_table_{uuid.uuid4()}"
        _set_up_delta_table(delta_table_path)

        total_row_number = _get_total_row_number(
            delta_table_path, pendulum.period(pendulum.parse("2023-09-17"), pendulum.parse("2023-09-19"))
        )

        assert total_row_number == 13

        _tear_down_delta_table(delta_table_path)

    def test_calculate_split_count(self):
        delta_table_path = f"/tmp/test_spine_split/delta_table_{uuid.uuid4()}"
        _set_up_delta_table(delta_table_path)

        split_count = _calculate_split_count(
            delta_table_path, pendulum.period(pendulum.parse("2023-09-17"), pendulum.parse("2023-09-19")), 4
        )

        assert split_count == 4

        _tear_down_delta_table(delta_table_path)


def _set_up_delta_table(delta_table_path):
    # fmt: off
    df = pandas.DataFrame(
        {
            "value": [
                1, 2, 3,
                1, 2,
                1, 2, 3, 4, 5, 6, 7,
                1, 2, 3, 4,
                1, 2,
                1, 2,
            ],
            "time_partition": [
                "2023-09-15", "2023-09-15", "2023-09-15",
                "2023-09-16", "2023-09-16",
                "2023-09-17", "2023-09-17", "2023-09-17", "2023-09-17", "2023-09-17", "2023-09-17", "2023-09-17",
                "2023-09-18", "2023-09-18", "2023-09-18", "2023-09-18",
                "2023-09-19", "2023-09-19",
                "2023-09-20", "2023-09-20",
            ],
        }
    )
    # fmt: on
    write_deltalake(delta_table_path, df, partition_by=["time_partition"])


def _tear_down_delta_table(delta_table_path):
    if os.path.exists(delta_table_path):
        shutil.rmtree(delta_table_path)
        print(f"Deleted Delta table at: {delta_table_path}")



import datetime
from collections import defaultdict
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import attrs

import tecton_core.tecton_pendulum as pendulum
from tecton_core import feature_definition_wrapper
from tecton_core import query_consts
from tecton_core import schema
from tecton_core import time_utils
from tecton_core.specs import LifetimeWindowSpec
from tecton_core.specs import OnlineBatchTablePart as SpecOnlineBatchTablePart
from tecton_core.specs import OnlineBatchTablePartTile as SpecOnlineBatchTablePartTile
from tecton_core.specs import TimeWindowSpec
from tecton_core.specs import create_time_window_spec_from_data_proto
from tecton_proto.data import feature_view__client_pb2 as feature_view__data_pb2


@attrs.frozen
class AggregationGroup:
    """AggregationGroup represents a group of aggregate features to compute with a corresponding start/end.

    The typical usage of this will be in compaction jobs, where we will use the start/end time to determine
    eligible rows for each individual aggregate. The tile index and start/end time correspond to an OnlineBatchTablePartTile for this aggregation window group. Each tile may represent a smaller time window, within the larger aggregation window, to compute futures for.
    """

    window_index: int
    inclusive_start_time: Optional[datetime.datetime]
    exclusive_end_time: datetime.datetime
    aggregate_features: Tuple[feature_view__data_pb2.Aggregate, ...]
    schema: schema.Schema
    tile_index: int
    window_tile_column_name: str  # 0_0 for example


def _get_inclusive_start_time_for_window(
    exclusive_end_time: datetime.datetime,
    aggregation_window: TimeWindowSpec,
    tile_window: Optional[SpecOnlineBatchTablePartTile] = None,
) -> Optional[datetime.datetime]:
    if isinstance(aggregation_window, LifetimeWindowSpec):
        return None
    window_start_time = (
        tile_window.relative_start_time_inclusive if tile_window is not None else aggregation_window.window_start
    )
    return time_utils.get_timezone_aware_datetime(exclusive_end_time + window_start_time)


def _get_exclusive_end_time_for_window(
    exclusive_end_time: datetime.datetime,
    aggregation_window: TimeWindowSpec,
    tile_window: Optional[SpecOnlineBatchTablePartTile] = None,
) -> datetime.datetime:
    if isinstance(aggregation_window, LifetimeWindowSpec):
        return time_utils.get_timezone_aware_datetime(exclusive_end_time)
    window_end_time = (
        tile_window.relative_end_time_exclusive if tile_window is not None else aggregation_window.window_end
    )
    return time_utils.get_timezone_aware_datetime(exclusive_end_time + window_end_time)


def _get_groups_for_aggregation_part(
    agg_part: SpecOnlineBatchTablePart, exclusive_end_time: datetime.datetime, aggregation_map: Dict
) -> List[AggregationGroup]:
    """Create one aggregation group for each tile in each aggregation part.

    Stream FVs with non-lifetime windows use sawtooths and have multiple tiles per part.
    Batch feature views with non-lifetime windows have one tile set per part.
    Lifetime windows do not have tiles set in their OnlineBatchTablePart, but we still add 1 agg group per part.
    """
    tiles = agg_part.tiles or [None]
    agg_group_tiles = []
    for i, tile in enumerate(tiles):
        agg_group_tiles.append(
            AggregationGroup(
                window_index=agg_part.window_index,
                inclusive_start_time=_get_inclusive_start_time_for_window(
                    exclusive_end_time, agg_part.time_window, tile
                ),
                exclusive_end_time=_get_exclusive_end_time_for_window(exclusive_end_time, agg_part.time_window, tile),
                aggregate_features=tuple(aggregation_map[agg_part.time_window]),
                schema=agg_part.schema,
                tile_index=i,
                window_tile_column_name=f"{agg_part.window_index}_{i}",
            )
        )
    return agg_group_tiles


def aggregation_groups(
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    exclusive_end_time: datetime.datetime,
) -> Tuple[AggregationGroup, ...]:
    aggregation_map = defaultdict(list)
    for aggregation in fdw.trailing_time_window_aggregation().features:
        aggregation_map[create_time_window_spec_from_data_proto(aggregation.time_window)].append(aggregation)

    aggregation_parts = fdw.fv_spec.online_batch_table_format.online_batch_table_parts

    if len(aggregation_parts) != len(aggregation_map):
        msg = "unexpected difference in length of the spec's online_batch_table_format and trailing_time_window_aggregation"
        raise ValueError(msg)

    agg_groups = []
    for agg_part in aggregation_parts:
        agg_groups += _get_groups_for_aggregation_part(agg_part, exclusive_end_time, aggregation_map)

    return tuple(agg_groups)


def _get_min_window_start_time(
    aggregation_groups: Tuple[AggregationGroup, ...], fdw: feature_definition_wrapper.FeatureDefinitionWrapper
) -> Optional[pendulum.DateTime]:
    contains_lifetime_agg = any(group.inclusive_start_time is None for group in aggregation_groups)
    if contains_lifetime_agg:
        return fdw.materialization_start_timestamp
    min_window_time = min(group.inclusive_start_time for group in aggregation_groups)
    return pendulum.instance(min_window_time)


def _get_max_window_end_time(aggregation_groups: Tuple[AggregationGroup, ...]) -> pendulum.DateTime:
    max_window_time = max(group.exclusive_end_time for group in aggregation_groups)
    return pendulum.instance(max_window_time)


def get_data_time_limits_for_compaction(
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper, compaction_job_end_time: datetime.datetime
) -> Optional[pendulum.Period]:
    """Compute the time filter to be used for online compaction jobs.

    This determines how much data to read from the offline store.
    For aggregate fvs,
        start_time=earliest agg window start
        end_time=latest agg window end
    For non agg fvs,
        start_time=max(feature start time, compaction_job_end_time - ttl)
        end_time=compaction_job_end_time"""
    if fdw.materialization_start_timestamp is None:
        return None

    if fdw.is_temporal_aggregate:
        agg_groups = aggregation_groups(fdw=fdw, exclusive_end_time=compaction_job_end_time)
        start_time = _get_min_window_start_time(agg_groups, fdw)
        end_time = _get_max_window_end_time(agg_groups)
        return pendulum.Period(start_time, end_time)

    if not fdw.is_temporal:
        msg = "Expected fv to be of type temporal or temporal aggregate."
        raise Exception(msg)

    # respect ttl and feature start time for temporal fvs
    end_time = pendulum.instance(compaction_job_end_time)
    if fdw.serving_ttl:
        if not fdw.feature_start_timestamp:
            msg = "Expected feature start time to be set for temporal fvs when ttl is set."
            raise Exception(msg)
        job_time_minus_ttl = end_time - fdw.serving_ttl
        start_time = max(fdw.feature_start_timestamp, job_time_minus_ttl)
    elif fdw.feature_start_timestamp:
        start_time = fdw.feature_start_timestamp
    else:
        msg = "Expected ttl or feature start time to be set for temporal fvs."
        raise Exception(msg)
    return pendulum.Period(start_time, end_time)


def get_sorted_tile_column_names(agg_window_index: int, agg_groups: List[AggregationGroup]) -> List[str]:
    """Calculate the list of window_tile_column_names in ascending order for a given aggregation window index."""
    assert all(
        agg_window_index == agg_group.window_index for agg_group in agg_groups
    ), "All aggregation groups must have the same window index."
    expected_tile_indexes = set(range(len(agg_groups)))
    assert {
        agg_group.tile_index for agg_group in agg_groups
    } == expected_tile_indexes, "All aggregation groups must have unique tile indexes."

    sorted_tiles = sorted(
        agg_groups,
        key=lambda agg_group: agg_group.tile_index,
    )
    return [tile.window_tile_column_name for tile in sorted_tiles]


def _get_anchor_time_column_for_sawtooth_size(sawtooth_size: Optional[pendulum.Duration]) -> str:
    if sawtooth_size == pendulum.Duration(days=1):
        return query_consts.anchor_time_for_day_sawtooth()
    elif sawtooth_size == pendulum.Duration(hours=1):
        return query_consts.anchor_time_for_hour_sawtooth()
    elif sawtooth_size is None or sawtooth_size < pendulum.Duration(hours=1):
        return query_consts.anchor_time_for_non_sawtooth()

    msg = f"Invalid sawtooth size: {sawtooth_size}"
    raise ValueError(msg)


def _get_boolean_partition_column_for_sawtooth_size(
    sawtooth_size: Optional[pendulum.Duration],
) -> str:
    if sawtooth_size == pendulum.Duration(days=1):
        return query_consts.is_day_sawtooth()
    elif sawtooth_size == pendulum.Duration(hours=1):
        return query_consts.is_hour_sawtooth()
    elif sawtooth_size is None or sawtooth_size < pendulum.Duration(hours=1):
        return query_consts.is_non_sawtooth()

    msg = f"Invalid sawtooth size: {sawtooth_size}"
    raise ValueError(msg)


def _get_time_partition_column_for_sawtooth_size(sawtooth_size: Optional[pendulum.Duration]) -> Optional[str]:
    if sawtooth_size is None:
        return None
    elif sawtooth_size == pendulum.Duration(days=1):
        return query_consts.time_partition_column_for_date()
    elif sawtooth_size == pendulum.Duration(hours=1):
        return query_consts.time_partition_column_for_hour()
    elif sawtooth_size < pendulum.Duration(hours=1):
        # Either 5 minutes or one minute.
        return query_consts.time_partition_column_for_minute()

    msg = f"Invalid sawtooth size: {sawtooth_size}"
    raise ValueError(msg)


@attrs.frozen
class SawtoothAggregateFeatureBase:
    """Base class for SawtoothAggregateFeature and SecondaryKeySawtoothAggregateFeature.

    anchor_time_column: The anchor time column to use for aggregating.
    identifier_partition_column: A column to partition the data by when aggregating. This column represents which sawtooth size to use for the aggregation time window.
    use_continuous_range_query: Whether to use a continuous range query for the aggregation instead of the sawtooth aggregation plan. Only true for <2 day continuous time windows.
    time_partition_column: A column to partition the data by when aggregating. This column represents the date/hour/5min/1min interval the anchor time column falls into.
    """

    anchor_time_column: str
    identifier_partition_column: str
    use_continuous_range_query: bool
    time_partition_column: Optional[str]


@attrs.frozen
class SawtoothAggregateFeature(SawtoothAggregateFeatureBase):
    """SawtoothAggregateFeature represents a single aggregate feature and its metadata.

    The typical usage of this will be in offline retrieval for Feature Views that use sawtooth aggregations (i.e. Stream Fvs with compaction enabled and non lifetime windows).
    """

    aggregate_feature: feature_view__data_pb2.Aggregate
    batch_sawtooth_tile_size: Optional[pendulum.Duration]


@attrs.frozen
class SecondaryKeySawtoothAggregateFeature(SawtoothAggregateFeatureBase):
    """SecondaryKeySawtoothAggregateFeature represents a single SecondaryKeyOutputColumn and its metadata.

    The typical usage of this will be in offline retrieval for Feature Views that use sawtooth aggregations and secondary key aggregations.(i.e. Stream Fvs with compaction enabled and non lifetime windows).
    """

    name: str


@attrs.frozen
class SawtoothAggregationData:
    """SawtoothAggregationData represents a collection of sawtooth aggregate features and their metadata.

    The typical usage of this will be in offline retrieval for Feature Views that use sawtooth aggregations (i.e. Stream Fvs with compaction enabled and non lifetime windows).
    """

    aggregations: Dict[str, SawtoothAggregateFeature]
    unique_batch_sawtooth_tile_sizes: List[pendulum.Duration]
    _secondary_key_aggregations: Dict[str, SecondaryKeySawtoothAggregateFeature]
    stream_tile_size: Optional[pendulum.Duration] = None

    @classmethod
    def from_aggregate_features(
        cls, fdw: feature_definition_wrapper.FeatureDefinitionWrapper
    ) -> "SawtoothAggregationData":
        sawtooth_aggregations = {}
        for agg in fdw.trailing_time_window_aggregation().features:
            batch_sawtooth_tile_size = (
                time_utils.proto_to_duration(agg.batch_sawtooth_tile_size)
                if agg.HasField("batch_sawtooth_tile_size")
                else None
            )
            batch_sawtooth_tile_size_for_retrieval = batch_sawtooth_tile_size
            if batch_sawtooth_tile_size is None:
                if agg.time_window.HasField("lifetime_window"):
                    # Lifetime windows do not use sawtooths online, but should mimic day sawtoothing behavior offline.
                    batch_sawtooth_tile_size_for_retrieval = pendulum.Duration(days=1)
                elif fdw.stream_tiling_enabled:
                    # For time windows < 2d, we use the stream tile size as the sawtooth size.
                    batch_sawtooth_tile_size_for_retrieval = fdw.stream_tile_size

            # If we support custom sawtooth sizes, we should move away from using hardcoded column names here.
            anchor_time_column = _get_anchor_time_column_for_sawtooth_size(batch_sawtooth_tile_size_for_retrieval)
            boolean_partition_column = _get_boolean_partition_column_for_sawtooth_size(
                batch_sawtooth_tile_size_for_retrieval
            )
            time_partition_column = _get_time_partition_column_for_sawtooth_size(batch_sawtooth_tile_size_for_retrieval)
            sawtooth_agg = SawtoothAggregateFeature(
                aggregate_feature=agg,
                batch_sawtooth_tile_size=batch_sawtooth_tile_size_for_retrieval,
                anchor_time_column=anchor_time_column,
                identifier_partition_column=boolean_partition_column,
                time_partition_column=time_partition_column,
                use_continuous_range_query=batch_sawtooth_tile_size_for_retrieval is None,
            )
            sawtooth_aggregations[agg.output_feature_name] = sawtooth_agg

        secondary_key_aggregations = {}
        if fdw.aggregation_secondary_key:
            for secondary_key_output in fdw.materialized_fv_spec.secondary_key_rollup_outputs:
                matching_agg_list = [
                    sawtooth_agg
                    for sawtooth_agg in sawtooth_aggregations.values()
                    if create_time_window_spec_from_data_proto(sawtooth_agg.aggregate_feature.time_window)
                    == secondary_key_output.time_window
                ]
                if len(matching_agg_list) == 0:
                    msg = f"Could not find an aggregation matching the time window for secondary key output {secondary_key_output.name}"
                    raise ValueError(msg)
                matching_agg = matching_agg_list[0]
                secondary_sawtooth_agg = SecondaryKeySawtoothAggregateFeature(
                    name=secondary_key_output.name,
                    anchor_time_column=matching_agg.anchor_time_column,
                    identifier_partition_column=matching_agg.identifier_partition_column,
                    time_partition_column=matching_agg.time_partition_column,
                    use_continuous_range_query=matching_agg.use_continuous_range_query,
                )
                secondary_key_aggregations[secondary_key_output.name] = secondary_sawtooth_agg

        return cls(
            aggregations=sawtooth_aggregations,
            unique_batch_sawtooth_tile_sizes=list(
                {agg.batch_sawtooth_tile_size for agg in sawtooth_aggregations.values()}
            ),
            stream_tile_size=fdw.stream_tile_size,
            secondary_key_aggregations=secondary_key_aggregations,
        )

    def get_sawtooth_aggregation_for_output_feature_name(self, output_feature_name: str) -> SawtoothAggregateFeature:
        if output_feature_name not in self.aggregations:
            msg = f"Feature output name {output_feature_name} not found in sawtooth aggregations."
            raise ValueError(msg)
        return self.aggregations[output_feature_name]

    def get_sawtooth_aggregate_feature_or_secondary_key_feature(
        self, name: str
    ) -> Union[SawtoothAggregateFeature, SecondaryKeySawtoothAggregateFeature]:
        if name in self.aggregations:
            return self.aggregations[name]
        elif name in self._secondary_key_aggregations:
            return self._secondary_key_aggregations[name]
        else:
            msg = f"Feature name {name} not found in sawtooth aggregations or secondary key sawtooth aggregations."
            raise ValueError(msg)

    def contains_day_sawtooths(self) -> bool:
        return any(tile_size == pendulum.Duration(days=1) for tile_size in self.unique_batch_sawtooth_tile_sizes)

    def contains_hour_sawtooths(self) -> bool:
        return any(tile_size == pendulum.Duration(hours=1) for tile_size in self.unique_batch_sawtooth_tile_sizes)

    def contains_non_sawtooth_aggregations(self) -> bool:
        # Aggregations < 2 days. Exception is  <2 day aggregations that have a 1hr stream tile size.
        return any(
            tile_size is None or tile_size < pendulum.Duration(hours=1)
            for tile_size in self.unique_batch_sawtooth_tile_sizes
        )

    def contains_continuous_non_sawtooth_aggregations(self) -> bool:
        return self.stream_tile_size is None and self.contains_non_sawtooth_aggregations()

    def contains_tiled_non_sawtooth_aggregations(self) -> bool:
        return self.stream_tile_size is not None and self.contains_non_sawtooth_aggregations()

    def get_anchor_time_columns(self, include_non_sawtooths: bool = True) -> List[str]:
        column_set = {
            agg.anchor_time_column
            for agg in self.aggregations.values()
            if include_non_sawtooths or agg.anchor_time_column != query_consts.anchor_time_for_non_sawtooth()
        }
        return list(column_set)

    def get_identifier_partition_columns(self) -> List[str]:
        column_set = {agg.identifier_partition_column for agg in self.aggregations.values()}
        return list(column_set)

    def get_truncated_timestamp_partition_columns(self) -> List[str]:
        column_set = {
            agg.time_partition_column for agg in self.aggregations.values() if agg.time_partition_column is not None
        }
        return list(column_set)

    def get_partition_column_to_bool_map(
        self, sawtooth_bool_value: bool = True, non_sawtooth_bool_value: bool = True
    ) -> Dict[str, bool]:
        # Map the new partition column names to the boolean constant the column should contain.
        partition_column_to_bool_map = {}
        if self.contains_day_sawtooths():
            partition_column_to_bool_map[query_consts.is_day_sawtooth()] = sawtooth_bool_value
        if self.contains_hour_sawtooths():
            partition_column_to_bool_map[query_consts.is_hour_sawtooth()] = sawtooth_bool_value
        if self.contains_non_sawtooth_aggregations():
            partition_column_to_bool_map[query_consts.is_non_sawtooth()] = non_sawtooth_bool_value
        return partition_column_to_bool_map

    def get_anchor_time_to_partition_columns_map(self, include_non_sawtooths: bool = True) -> Dict[str, str]:
        # Map the anchor time column used for aggregations to the column to partition by.
        anchor_time_to_identifier_map = {}
        if self.contains_day_sawtooths():
            anchor_time_to_identifier_map[query_consts.anchor_time_for_day_sawtooth()] = query_consts.is_day_sawtooth()
        if self.contains_hour_sawtooths():
            anchor_time_to_identifier_map[query_consts.anchor_time_for_hour_sawtooth()] = (
                query_consts.is_hour_sawtooth()
            )
        if include_non_sawtooths and self.contains_non_sawtooth_aggregations():
            anchor_time_to_identifier_map[query_consts.anchor_time_for_non_sawtooth()] = query_consts.is_non_sawtooth()
        return anchor_time_to_identifier_map

    def get_anchor_time_to_timedelta_map(self, use_zero_timedelta: bool = False) -> Dict[str, datetime.timedelta]:
        # Map the new anchor time column names to the duration of time the timestamp is rounded to.
        anchor_time_to_timedelta_map = {}
        if self.contains_day_sawtooths():
            day_duration = datetime.timedelta(seconds=0) if use_zero_timedelta else datetime.timedelta(days=1)
            anchor_time_to_timedelta_map[query_consts.anchor_time_for_day_sawtooth()] = day_duration
        if self.contains_hour_sawtooths():
            hour_duration = datetime.timedelta(seconds=0) if use_zero_timedelta else datetime.timedelta(hours=1)
            anchor_time_to_timedelta_map[query_consts.anchor_time_for_hour_sawtooth()] = hour_duration
        if self.contains_non_sawtooth_aggregations():
            if self.stream_tile_size:
                stream_tile_size_duration = (
                    datetime.timedelta(seconds=0) if use_zero_timedelta else self.stream_tile_size
                )
                anchor_time_to_timedelta_map[query_consts.anchor_time_for_non_sawtooth()] = stream_tile_size_duration
            else:
                anchor_time_to_timedelta_map[query_consts.anchor_time_for_non_sawtooth()] = datetime.timedelta(
                    seconds=0
                )
        return anchor_time_to_timedelta_map

    def get_anchor_time_to_aggregation_interval_map(
        self, aggregation_tile_interval: int, feature_store_format_version: int
    ) -> Dict[str, int]:
        # Map the new anchor time column names to aggregation tile interval the relevant partial tiles use.
        anchor_time_to_tile_interval_map = {}
        if self.contains_day_sawtooths():
            anchor_time_to_tile_interval_map[query_consts.anchor_time_for_day_sawtooth()] = aggregation_tile_interval
        if self.contains_hour_sawtooths():
            anchor_time_to_tile_interval_map[query_consts.anchor_time_for_hour_sawtooth()] = aggregation_tile_interval
        if self.contains_non_sawtooth_aggregations():
            if self.stream_tile_size:
                tile_interval = time_utils.convert_timedelta_for_version(
                    self.stream_tile_size, feature_store_format_version
                )
            else:
                tile_interval = 0
            anchor_time_to_tile_interval_map[query_consts.anchor_time_for_non_sawtooth()] = tile_interval
        return anchor_time_to_tile_interval_map



from tecton_proto.spark_common import clusters__client_pb2 as _clusters__client_pb2
from tecton_proto.spark_common import libraries__client_pb2 as _libraries__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AccessControlList(_message.Message):
    __slots__ = ["group_name", "permission_level", "service_principal_name", "user_name"]
    class PermissionLevel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    CAN_MANAGE: AccessControlList.PermissionLevel
    CAN_MANAGE_RUN: AccessControlList.PermissionLevel
    CAN_VIEW: AccessControlList.PermissionLevel
    GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    IS_OWNER: AccessControlList.PermissionLevel
    PERMISSION_LEVEL_FIELD_NUMBER: _ClassVar[int]
    SERVICE_PRINCIPAL_NAME_FIELD_NUMBER: _ClassVar[int]
    USER_NAME_FIELD_NUMBER: _ClassVar[int]
    group_name: str
    permission_level: AccessControlList.PermissionLevel
    service_principal_name: str
    user_name: str
    def __init__(self, user_name: _Optional[str] = ..., group_name: _Optional[str] = ..., service_principal_name: _Optional[str] = ..., permission_level: _Optional[_Union[AccessControlList.PermissionLevel, str]] = ...) -> None: ...

class ClusterInstance(_message.Message):
    __slots__ = ["cluster_id"]
    CLUSTER_ID_FIELD_NUMBER: _ClassVar[int]
    cluster_id: str
    def __init__(self, cluster_id: _Optional[str] = ...) -> None: ...

class ClusterSpec(_message.Message):
    __slots__ = ["existing_cluster_id", "libraries", "new_cluster"]
    EXISTING_CLUSTER_ID_FIELD_NUMBER: _ClassVar[int]
    LIBRARIES_FIELD_NUMBER: _ClassVar[int]
    NEW_CLUSTER_FIELD_NUMBER: _ClassVar[int]
    existing_cluster_id: str
    libraries: _containers.RepeatedCompositeFieldContainer[RemoteLibrary]
    new_cluster: NewCluster
    def __init__(self, new_cluster: _Optional[_Union[NewCluster, _Mapping]] = ..., existing_cluster_id: _Optional[str] = ..., libraries: _Optional[_Iterable[_Union[RemoteLibrary, _Mapping]]] = ...) -> None: ...

class JobsCancelRunRequest(_message.Message):
    __slots__ = ["run_id"]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    run_id: int
    def __init__(self, run_id: _Optional[int] = ...) -> None: ...

class JobsRunsGetRequest(_message.Message):
    __slots__ = ["run_id"]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    run_id: int
    def __init__(self, run_id: _Optional[int] = ...) -> None: ...

class JobsRunsGetResponse(_message.Message):
    __slots__ = ["cluster_instance", "end_time", "execution_duration", "job_id", "run_id", "run_page_url", "setup_duration", "start_time", "state"]
    CLUSTER_INSTANCE_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_DURATION_FIELD_NUMBER: _ClassVar[int]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    RUN_PAGE_URL_FIELD_NUMBER: _ClassVar[int]
    SETUP_DURATION_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    cluster_instance: ClusterInstance
    end_time: int
    execution_duration: int
    job_id: int
    run_id: int
    run_page_url: str
    setup_duration: int
    start_time: int
    state: RunState
    def __init__(self, run_id: _Optional[int] = ..., job_id: _Optional[int] = ..., execution_duration: _Optional[int] = ..., start_time: _Optional[int] = ..., end_time: _Optional[int] = ..., setup_duration: _Optional[int] = ..., cluster_instance: _Optional[_Union[ClusterInstance, _Mapping]] = ..., run_page_url: _Optional[str] = ..., state: _Optional[_Union[RunState, _Mapping]] = ...) -> None: ...

class JobsRunsListRequest(_message.Message):
    __slots__ = ["active_only", "limit", "offset", "run_type"]
    ACTIVE_ONLY_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    RUN_TYPE_FIELD_NUMBER: _ClassVar[int]
    active_only: bool
    limit: int
    offset: int
    run_type: str
    def __init__(self, offset: _Optional[int] = ..., active_only: bool = ..., run_type: _Optional[str] = ..., limit: _Optional[int] = ...) -> None: ...

class JobsRunsListResponse(_message.Message):
    __slots__ = ["has_more", "runs"]
    HAS_MORE_FIELD_NUMBER: _ClassVar[int]
    RUNS_FIELD_NUMBER: _ClassVar[int]
    has_more: bool
    runs: _containers.RepeatedCompositeFieldContainer[Run]
    def __init__(self, runs: _Optional[_Iterable[_Union[Run, _Mapping]]] = ..., has_more: bool = ...) -> None: ...

class JobsRunsSubmitRequest(_message.Message):
    __slots__ = ["access_control_list", "idempotency_token", "run_as", "run_name", "tasks", "timeout_seconds"]
    ACCESS_CONTROL_LIST_FIELD_NUMBER: _ClassVar[int]
    IDEMPOTENCY_TOKEN_FIELD_NUMBER: _ClassVar[int]
    RUN_AS_FIELD_NUMBER: _ClassVar[int]
    RUN_NAME_FIELD_NUMBER: _ClassVar[int]
    TASKS_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_SECONDS_FIELD_NUMBER: _ClassVar[int]
    access_control_list: _containers.RepeatedCompositeFieldContainer[AccessControlList]
    idempotency_token: str
    run_as: RunAs
    run_name: str
    tasks: _containers.RepeatedCompositeFieldContainer[Task]
    timeout_seconds: int
    def __init__(self, run_name: _Optional[str] = ..., timeout_seconds: _Optional[int] = ..., idempotency_token: _Optional[str] = ..., tasks: _Optional[_Iterable[_Union[Task, _Mapping]]] = ..., access_control_list: _Optional[_Iterable[_Union[AccessControlList, _Mapping]]] = ..., run_as: _Optional[_Union[RunAs, _Mapping]] = ...) -> None: ...

class JobsRunsSubmitResponse(_message.Message):
    __slots__ = ["run_id"]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    run_id: int
    def __init__(self, run_id: _Optional[int] = ...) -> None: ...

class LegacyJobsRunsSubmitRequest(_message.Message):
    __slots__ = ["existing_cluster_id", "libraries", "new_cluster", "notebook_task", "run_name", "tasks", "timeout_seconds"]
    EXISTING_CLUSTER_ID_FIELD_NUMBER: _ClassVar[int]
    LIBRARIES_FIELD_NUMBER: _ClassVar[int]
    NEW_CLUSTER_FIELD_NUMBER: _ClassVar[int]
    NOTEBOOK_TASK_FIELD_NUMBER: _ClassVar[int]
    RUN_NAME_FIELD_NUMBER: _ClassVar[int]
    TASKS_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_SECONDS_FIELD_NUMBER: _ClassVar[int]
    existing_cluster_id: str
    libraries: _containers.RepeatedCompositeFieldContainer[RemoteLibrary]
    new_cluster: _clusters__client_pb2.NewCluster
    notebook_task: NotebookTask
    run_name: str
    tasks: _containers.RepeatedCompositeFieldContainer[Task]
    timeout_seconds: int
    def __init__(self, new_cluster: _Optional[_Union[_clusters__client_pb2.NewCluster, _Mapping]] = ..., existing_cluster_id: _Optional[str] = ..., notebook_task: _Optional[_Union[NotebookTask, _Mapping]] = ..., run_name: _Optional[str] = ..., libraries: _Optional[_Iterable[_Union[RemoteLibrary, _Mapping]]] = ..., timeout_seconds: _Optional[int] = ..., tasks: _Optional[_Iterable[_Union[Task, _Mapping]]] = ...) -> None: ...

class NewCluster(_message.Message):
    __slots__ = ["custom_tags"]
    class CustomTagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    CUSTOM_TAGS_FIELD_NUMBER: _ClassVar[int]
    custom_tags: _containers.ScalarMap[str, str]
    def __init__(self, custom_tags: _Optional[_Mapping[str, str]] = ...) -> None: ...

class NotebookTask(_message.Message):
    __slots__ = ["base_parameters", "notebook_path"]
    class BaseParametersEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    BASE_PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    NOTEBOOK_PATH_FIELD_NUMBER: _ClassVar[int]
    base_parameters: _containers.ScalarMap[str, str]
    notebook_path: str
    def __init__(self, notebook_path: _Optional[str] = ..., base_parameters: _Optional[_Mapping[str, str]] = ...) -> None: ...

class RemoteLibrary(_message.Message):
    __slots__ = ["egg", "jar", "maven", "pypi", "whl"]
    EGG_FIELD_NUMBER: _ClassVar[int]
    JAR_FIELD_NUMBER: _ClassVar[int]
    MAVEN_FIELD_NUMBER: _ClassVar[int]
    PYPI_FIELD_NUMBER: _ClassVar[int]
    WHL_FIELD_NUMBER: _ClassVar[int]
    egg: str
    jar: str
    maven: _libraries__client_pb2.MavenLibrary
    pypi: _libraries__client_pb2.PyPiLibrary
    whl: str
    def __init__(self, jar: _Optional[str] = ..., egg: _Optional[str] = ..., whl: _Optional[str] = ..., maven: _Optional[_Union[_libraries__client_pb2.MavenLibrary, _Mapping]] = ..., pypi: _Optional[_Union[_libraries__client_pb2.PyPiLibrary, _Mapping]] = ...) -> None: ...

class Run(_message.Message):
    __slots__ = ["cluster_spec", "job_id", "run_id", "run_page_url", "state"]
    CLUSTER_SPEC_FIELD_NUMBER: _ClassVar[int]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    RUN_PAGE_URL_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    cluster_spec: ClusterSpec
    job_id: int
    run_id: int
    run_page_url: str
    state: RunState
    def __init__(self, job_id: _Optional[int] = ..., run_id: _Optional[int] = ..., state: _Optional[_Union[RunState, _Mapping]] = ..., cluster_spec: _Optional[_Union[ClusterSpec, _Mapping]] = ..., run_page_url: _Optional[str] = ...) -> None: ...

class RunAs(_message.Message):
    __slots__ = ["service_principal_name", "user_name"]
    SERVICE_PRINCIPAL_NAME_FIELD_NUMBER: _ClassVar[int]
    USER_NAME_FIELD_NUMBER: _ClassVar[int]
    service_principal_name: str
    user_name: str
    def __init__(self, user_name: _Optional[str] = ..., service_principal_name: _Optional[str] = ...) -> None: ...

class RunState(_message.Message):
    __slots__ = ["life_cycle_state", "result_state", "state_message"]
    class RunLifeCycleState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class RunResultState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    CANCELED: RunState.RunResultState
    FAILED: RunState.RunResultState
    INTERNAL_ERROR: RunState.RunLifeCycleState
    LIFE_CYCLE_STATE_FIELD_NUMBER: _ClassVar[int]
    PENDING: RunState.RunLifeCycleState
    RESULT_STATE_FIELD_NUMBER: _ClassVar[int]
    RUNNING: RunState.RunLifeCycleState
    SKIPPED: RunState.RunLifeCycleState
    STATE_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    SUCCESS: RunState.RunResultState
    TERMINATED: RunState.RunLifeCycleState
    TERMINATING: RunState.RunLifeCycleState
    TIMEDOUT: RunState.RunResultState
    UNKNOWN_RUN_LIFE_CYCLE_STATE: RunState.RunLifeCycleState
    UNKNOWN_RUN_RESULT_STATE: RunState.RunResultState
    life_cycle_state: RunState.RunLifeCycleState
    result_state: RunState.RunResultState
    state_message: str
    def __init__(self, life_cycle_state: _Optional[_Union[RunState.RunLifeCycleState, str]] = ..., result_state: _Optional[_Union[RunState.RunResultState, str]] = ..., state_message: _Optional[str] = ...) -> None: ...

class Task(_message.Message):
    __slots__ = ["depends_on", "description", "existing_cluster_id", "libraries", "new_cluster", "notebook_task", "task_key", "timeout_seconds"]
    DEPENDS_ON_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    EXISTING_CLUSTER_ID_FIELD_NUMBER: _ClassVar[int]
    LIBRARIES_FIELD_NUMBER: _ClassVar[int]
    NEW_CLUSTER_FIELD_NUMBER: _ClassVar[int]
    NOTEBOOK_TASK_FIELD_NUMBER: _ClassVar[int]
    TASK_KEY_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_SECONDS_FIELD_NUMBER: _ClassVar[int]
    depends_on: _containers.RepeatedCompositeFieldContainer[Task]
    description: str
    existing_cluster_id: str
    libraries: _containers.RepeatedCompositeFieldContainer[RemoteLibrary]
    new_cluster: _clusters__client_pb2.NewCluster
    notebook_task: NotebookTask
    task_key: str
    timeout_seconds: int
    def __init__(self, task_key: _Optional[str] = ..., description: _Optional[str] = ..., depends_on: _Optional[_Iterable[_Union[Task, _Mapping]]] = ..., new_cluster: _Optional[_Union[_clusters__client_pb2.NewCluster, _Mapping]] = ..., existing_cluster_id: _Optional[str] = ..., notebook_task: _Optional[_Union[NotebookTask, _Mapping]] = ..., libraries: _Optional[_Iterable[_Union[RemoteLibrary, _Mapping]]] = ..., timeout_seconds: _Optional[int] = ...) -> None: ...



from tecton_proto.common import aws_credentials__client_pb2 as _aws_credentials__client_pb2
from tecton_proto.common import secret__client_pb2 as _secret__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AwsSettings(_message.Message):
    __slots__ = ["compute_extra_tags", "dynamo_extra_tags", "dynamo_role", "dynamo_table_names", "ec2_settings", "elasticache_extra_tags", "elasticache_settings", "emr_settings", "object_store_locations"]
    class ComputeExtraTagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class DynamoExtraTagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class ElasticacheExtraTagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    COMPUTE_EXTRA_TAGS_FIELD_NUMBER: _ClassVar[int]
    DYNAMO_EXTRA_TAGS_FIELD_NUMBER: _ClassVar[int]
    DYNAMO_ROLE_FIELD_NUMBER: _ClassVar[int]
    DYNAMO_TABLE_NAMES_FIELD_NUMBER: _ClassVar[int]
    EC2_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    ELASTICACHE_EXTRA_TAGS_FIELD_NUMBER: _ClassVar[int]
    ELASTICACHE_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    EMR_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    OBJECT_STORE_LOCATIONS_FIELD_NUMBER: _ClassVar[int]
    compute_extra_tags: _containers.ScalarMap[str, str]
    dynamo_extra_tags: _containers.ScalarMap[str, str]
    dynamo_role: _aws_credentials__client_pb2.AwsIamRole
    dynamo_table_names: DynamoTableNames
    ec2_settings: Ec2Settings
    elasticache_extra_tags: _containers.ScalarMap[str, str]
    elasticache_settings: ElasticacheSettings
    emr_settings: EmrSettings
    object_store_locations: ObjectStoreLocations
    def __init__(self, dynamo_role: _Optional[_Union[_aws_credentials__client_pb2.AwsIamRole, _Mapping]] = ..., dynamo_extra_tags: _Optional[_Mapping[str, str]] = ..., compute_extra_tags: _Optional[_Mapping[str, str]] = ..., elasticache_extra_tags: _Optional[_Mapping[str, str]] = ..., emr_settings: _Optional[_Union[EmrSettings, _Mapping]] = ..., ec2_settings: _Optional[_Union[Ec2Settings, _Mapping]] = ..., elasticache_settings: _Optional[_Union[ElasticacheSettings, _Mapping]] = ..., dynamo_table_names: _Optional[_Union[DynamoTableNames, _Mapping]] = ..., object_store_locations: _Optional[_Union[ObjectStoreLocations, _Mapping]] = ...) -> None: ...

class DBFSLocation(_message.Message):
    __slots__ = ["path"]
    PATH_FIELD_NUMBER: _ClassVar[int]
    path: str
    def __init__(self, path: _Optional[str] = ...) -> None: ...

class DatabricksConfig(_message.Message):
    __slots__ = ["api_token", "spark_version", "user_display_name", "user_name", "workspace_url"]
    API_TOKEN_FIELD_NUMBER: _ClassVar[int]
    SPARK_VERSION_FIELD_NUMBER: _ClassVar[int]
    USER_DISPLAY_NAME_FIELD_NUMBER: _ClassVar[int]
    USER_NAME_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_URL_FIELD_NUMBER: _ClassVar[int]
    api_token: _secret__client_pb2.Secret
    spark_version: str
    user_display_name: str
    user_name: str
    workspace_url: str
    def __init__(self, workspace_url: _Optional[str] = ..., api_token: _Optional[_Union[_secret__client_pb2.Secret, _Mapping]] = ..., user_name: _Optional[str] = ..., user_display_name: _Optional[str] = ..., spark_version: _Optional[str] = ...) -> None: ...

class DatabricksWorkspaceFileLocation(_message.Message):
    __slots__ = ["path"]
    PATH_FIELD_NUMBER: _ClassVar[int]
    path: str
    def __init__(self, path: _Optional[str] = ...) -> None: ...

class DynamoTableNames(_message.Message):
    __slots__ = ["canary_table_name", "data_table_prefix", "delta_log_table_name", "delta_log_table_name_v2", "job_idempotence_key_table_name", "job_metadata_table_name", "metric_table_prefix", "status_table_name"]
    CANARY_TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    DATA_TABLE_PREFIX_FIELD_NUMBER: _ClassVar[int]
    DELTA_LOG_TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    DELTA_LOG_TABLE_NAME_V2_FIELD_NUMBER: _ClassVar[int]
    JOB_IDEMPOTENCE_KEY_TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    JOB_METADATA_TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    METRIC_TABLE_PREFIX_FIELD_NUMBER: _ClassVar[int]
    STATUS_TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    canary_table_name: str
    data_table_prefix: str
    delta_log_table_name: str
    delta_log_table_name_v2: str
    job_idempotence_key_table_name: str
    job_metadata_table_name: str
    metric_table_prefix: str
    status_table_name: str
    def __init__(self, data_table_prefix: _Optional[str] = ..., status_table_name: _Optional[str] = ..., job_idempotence_key_table_name: _Optional[str] = ..., canary_table_name: _Optional[str] = ..., delta_log_table_name: _Optional[str] = ..., metric_table_prefix: _Optional[str] = ..., delta_log_table_name_v2: _Optional[str] = ..., job_metadata_table_name: _Optional[str] = ...) -> None: ...

class Ec2Settings(_message.Message):
    __slots__ = ["ray_cluster_manager_role", "ray_instance_profile"]
    RAY_CLUSTER_MANAGER_ROLE_FIELD_NUMBER: _ClassVar[int]
    RAY_INSTANCE_PROFILE_FIELD_NUMBER: _ClassVar[int]
    ray_cluster_manager_role: _aws_credentials__client_pb2.AwsIamRole
    ray_instance_profile: _aws_credentials__client_pb2.AwsIamRole
    def __init__(self, ray_cluster_manager_role: _Optional[_Union[_aws_credentials__client_pb2.AwsIamRole, _Mapping]] = ..., ray_instance_profile: _Optional[_Union[_aws_credentials__client_pb2.AwsIamRole, _Mapping]] = ...) -> None: ...

class ElasticacheSettings(_message.Message):
    __slots__ = ["elasticache_manager_role", "elasticache_subnet_group"]
    ELASTICACHE_MANAGER_ROLE_FIELD_NUMBER: _ClassVar[int]
    ELASTICACHE_SUBNET_GROUP_FIELD_NUMBER: _ClassVar[int]
    elasticache_manager_role: _aws_credentials__client_pb2.AwsIamRole
    elasticache_subnet_group: str
    def __init__(self, elasticache_manager_role: _Optional[_Union[_aws_credentials__client_pb2.AwsIamRole, _Mapping]] = ..., elasticache_subnet_group: _Optional[str] = ...) -> None: ...

class EmrSettings(_message.Message):
    __slots__ = ["emr_control_role"]
    EMR_CONTROL_ROLE_FIELD_NUMBER: _ClassVar[int]
    emr_control_role: _aws_credentials__client_pb2.AwsIamRole
    def __init__(self, emr_control_role: _Optional[_Union[_aws_credentials__client_pb2.AwsIamRole, _Mapping]] = ...) -> None: ...

class GCSLocation(_message.Message):
    __slots__ = ["path"]
    PATH_FIELD_NUMBER: _ClassVar[int]
    path: str
    def __init__(self, path: _Optional[str] = ...) -> None: ...

class ObjectStoreLocation(_message.Message):
    __slots__ = ["dbfs_location", "gcs_location", "s3_location", "workspace_location"]
    DBFS_LOCATION_FIELD_NUMBER: _ClassVar[int]
    GCS_LOCATION_FIELD_NUMBER: _ClassVar[int]
    S3_LOCATION_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_LOCATION_FIELD_NUMBER: _ClassVar[int]
    dbfs_location: DBFSLocation
    gcs_location: GCSLocation
    s3_location: S3Location
    workspace_location: DatabricksWorkspaceFileLocation
    def __init__(self, s3_location: _Optional[_Union[S3Location, _Mapping]] = ..., dbfs_location: _Optional[_Union[DBFSLocation, _Mapping]] = ..., gcs_location: _Optional[_Union[GCSLocation, _Mapping]] = ..., workspace_location: _Optional[_Union[DatabricksWorkspaceFileLocation, _Mapping]] = ...) -> None: ...

class ObjectStoreLocations(_message.Message):
    __slots__ = ["custom_environment_dependencies", "data_validation", "databricks_scripts", "emr_scripts", "feature_export", "feature_repo", "feature_server_configuration", "feature_server_logging", "intermediate_data", "job_metadata_table", "kafka_credentials_base", "materialization", "materialization_params", "model_artifacts", "observability_service_configuration", "push_api_configuration", "rift_logs", "self_serve_consumption", "streaming_checkpoint", "system_audit_logging", "transform_server_group_configuration", "transformation_config"]
    CUSTOM_ENVIRONMENT_DEPENDENCIES_FIELD_NUMBER: _ClassVar[int]
    DATABRICKS_SCRIPTS_FIELD_NUMBER: _ClassVar[int]
    DATA_VALIDATION_FIELD_NUMBER: _ClassVar[int]
    EMR_SCRIPTS_FIELD_NUMBER: _ClassVar[int]
    FEATURE_EXPORT_FIELD_NUMBER: _ClassVar[int]
    FEATURE_REPO_FIELD_NUMBER: _ClassVar[int]
    FEATURE_SERVER_CONFIGURATION_FIELD_NUMBER: _ClassVar[int]
    FEATURE_SERVER_LOGGING_FIELD_NUMBER: _ClassVar[int]
    INTERMEDIATE_DATA_FIELD_NUMBER: _ClassVar[int]
    JOB_METADATA_TABLE_FIELD_NUMBER: _ClassVar[int]
    KAFKA_CREDENTIALS_BASE_FIELD_NUMBER: _ClassVar[int]
    MATERIALIZATION_FIELD_NUMBER: _ClassVar[int]
    MATERIALIZATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    MODEL_ARTIFACTS_FIELD_NUMBER: _ClassVar[int]
    OBSERVABILITY_SERVICE_CONFIGURATION_FIELD_NUMBER: _ClassVar[int]
    PUSH_API_CONFIGURATION_FIELD_NUMBER: _ClassVar[int]
    RIFT_LOGS_FIELD_NUMBER: _ClassVar[int]
    SELF_SERVE_CONSUMPTION_FIELD_NUMBER: _ClassVar[int]
    STREAMING_CHECKPOINT_FIELD_NUMBER: _ClassVar[int]
    SYSTEM_AUDIT_LOGGING_FIELD_NUMBER: _ClassVar[int]
    TRANSFORMATION_CONFIG_FIELD_NUMBER: _ClassVar[int]
    TRANSFORM_SERVER_GROUP_CONFIGURATION_FIELD_NUMBER: _ClassVar[int]
    custom_environment_dependencies: ObjectStoreLocation
    data_validation: ObjectStoreLocation
    databricks_scripts: ObjectStoreLocation
    emr_scripts: ObjectStoreLocation
    feature_export: ObjectStoreLocation
    feature_repo: ObjectStoreLocation
    feature_server_configuration: ObjectStoreLocation
    feature_server_logging: ObjectStoreLocation
    intermediate_data: ObjectStoreLocation
    job_metadata_table: ObjectStoreLocation
    kafka_credentials_base: ObjectStoreLocation
    materialization: ObjectStoreLocation
    materialization_params: ObjectStoreLocation
    model_artifacts: ObjectStoreLocation
    observability_service_configuration: ObjectStoreLocation
    push_api_configuration: ObjectStoreLocation
    rift_logs: ObjectStoreLocation
    self_serve_consumption: ObjectStoreLocation
    streaming_checkpoint: ObjectStoreLocation
    system_audit_logging: ObjectStoreLocation
    transform_server_group_configuration: ObjectStoreLocation
    transformation_config: ObjectStoreLocation
    def __init__(self, materialization: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., streaming_checkpoint: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., feature_server_configuration: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., feature_repo: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., emr_scripts: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., materialization_params: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., intermediate_data: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., feature_server_logging: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., kafka_credentials_base: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., job_metadata_table: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., push_api_configuration: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., data_validation: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., observability_service_configuration: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., system_audit_logging: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., databricks_scripts: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., self_serve_consumption: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., custom_environment_dependencies: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., feature_export: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., transformation_config: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., model_artifacts: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., rift_logs: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ..., transform_server_group_configuration: _Optional[_Union[ObjectStoreLocation, _Mapping]] = ...) -> None: ...

class S3Location(_message.Message):
    __slots__ = ["path", "role"]
    PATH_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    path: str
    role: _aws_credentials__client_pb2.AwsIamRole
    def __init__(self, path: _Optional[str] = ..., role: _Optional[_Union[_aws_credentials__client_pb2.AwsIamRole, _Mapping]] = ...) -> None: ...

class TenantSettingsProto(_message.Message):
    __slots__ = ["aws_settings", "base_feature_service_url", "base_metadata_service_url", "chronosphere_api_key", "chronosphere_restrict_label_value", "chronosphere_tecton_cluster_name", "customer_facing_tenant_name", "enable_user_editing_deployment_settings", "internal_tenant_name", "okta_user_group_id", "pseudonymize_amplitude_user_name", "spicedb_organization_name"]
    AWS_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    BASE_FEATURE_SERVICE_URL_FIELD_NUMBER: _ClassVar[int]
    BASE_METADATA_SERVICE_URL_FIELD_NUMBER: _ClassVar[int]
    CHRONOSPHERE_API_KEY_FIELD_NUMBER: _ClassVar[int]
    CHRONOSPHERE_RESTRICT_LABEL_VALUE_FIELD_NUMBER: _ClassVar[int]
    CHRONOSPHERE_TECTON_CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    CUSTOMER_FACING_TENANT_NAME_FIELD_NUMBER: _ClassVar[int]
    ENABLE_USER_EDITING_DEPLOYMENT_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    INTERNAL_TENANT_NAME_FIELD_NUMBER: _ClassVar[int]
    OKTA_USER_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    PSEUDONYMIZE_AMPLITUDE_USER_NAME_FIELD_NUMBER: _ClassVar[int]
    SPICEDB_ORGANIZATION_NAME_FIELD_NUMBER: _ClassVar[int]
    aws_settings: AwsSettings
    base_feature_service_url: str
    base_metadata_service_url: str
    chronosphere_api_key: _secret__client_pb2.Secret
    chronosphere_restrict_label_value: str
    chronosphere_tecton_cluster_name: str
    customer_facing_tenant_name: str
    enable_user_editing_deployment_settings: bool
    internal_tenant_name: str
    okta_user_group_id: str
    pseudonymize_amplitude_user_name: bool
    spicedb_organization_name: str
    def __init__(self, chronosphere_api_key: _Optional[_Union[_secret__client_pb2.Secret, _Mapping]] = ..., chronosphere_restrict_label_value: _Optional[str] = ..., pseudonymize_amplitude_user_name: bool = ..., enable_user_editing_deployment_settings: bool = ..., okta_user_group_id: _Optional[str] = ..., base_metadata_service_url: _Optional[str] = ..., base_feature_service_url: _Optional[str] = ..., spicedb_organization_name: _Optional[str] = ..., customer_facing_tenant_name: _Optional[str] = ..., chronosphere_tecton_cluster_name: _Optional[str] = ..., aws_settings: _Optional[_Union[AwsSettings, _Mapping]] = ..., internal_tenant_name: _Optional[str] = ...) -> None: ...

class UserDeploymentSettings(_message.Message):
    __slots__ = ["databricks_config", "tenant_settings", "user_deployment_settings_version", "user_spark_settings"]
    DATABRICKS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    TENANT_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    USER_DEPLOYMENT_SETTINGS_VERSION_FIELD_NUMBER: _ClassVar[int]
    USER_SPARK_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    databricks_config: DatabricksConfig
    tenant_settings: TenantSettingsProto
    user_deployment_settings_version: int
    user_spark_settings: UserSparkSettings
    def __init__(self, user_deployment_settings_version: _Optional[int] = ..., databricks_config: _Optional[_Union[DatabricksConfig, _Mapping]] = ..., user_spark_settings: _Optional[_Union[UserSparkSettings, _Mapping]] = ..., tenant_settings: _Optional[_Union[TenantSettingsProto, _Mapping]] = ...) -> None: ...

class UserSparkSettings(_message.Message):
    __slots__ = ["instance_profile_arn", "spark_conf"]
    class SparkConfEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    INSTANCE_PROFILE_ARN_FIELD_NUMBER: _ClassVar[int]
    SPARK_CONF_FIELD_NUMBER: _ClassVar[int]
    instance_profile_arn: str
    spark_conf: _containers.ScalarMap[str, str]
    def __init__(self, instance_profile_arn: _Optional[str] = ..., spark_conf: _Optional[_Mapping[str, str]] = ...) -> None: ...



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SparkField(_message.Message):
    __slots__ = ["name", "structfield_json"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    STRUCTFIELD_JSON_FIELD_NUMBER: _ClassVar[int]
    name: str
    structfield_json: str
    def __init__(self, name: _Optional[str] = ..., structfield_json: _Optional[str] = ...) -> None: ...

class SparkSchema(_message.Message):
    __slots__ = ["fields"]
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    fields: _containers.RepeatedCompositeFieldContainer[SparkField]
    def __init__(self, fields: _Optional[_Iterable[_Union[SparkField, _Mapping]]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/modelartifactservice/model_artifact_data__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.auth import principal__client_pb2 as tecton__proto_dot_auth_dot_principal____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nCtecton_proto/modelartifactservice/model_artifact_data__client.proto\x12!tecton_proto.modelartifactservice\x1a\x1fgoogle/protobuf/timestamp.proto\x1a)tecton_proto/auth/principal__client.proto\x1a$tecton_proto/common/id__client.proto\x1a(tecton_proto/common/schema__client.proto\"\x8c\t\n\x11ModelArtifactInfo\x12\'\n\x02id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12@\n\x04type\x18\x03 \x01(\x0e\x32,.tecton_proto.modelartifactservice.ModelTypeR\x04type\x12 \n\x0b\x64\x65scription\x18\x04 \x01(\tR\x0b\x64\x65scription\x12\x65\n\x0b\x66ile_hashes\x18\x05 \x03(\x0b\x32\x44.tecton_proto.modelartifactservice.ModelArtifactInfo.FileHashesEntryR\nfileHashes\x12>\n\x0cinput_schema\x18\x06 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x0binputSchema\x12@\n\routput_schema\x18\x07 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x0coutputSchema\x12R\n\x04tags\x18\t \x03(\x0b\x32>.tecton_proto.modelartifactservice.ModelArtifactInfo.TagsEntryR\x04tags\x12!\n\x0cstorage_path\x18\x08 \x01(\tR\x0bstoragePath\x12N\n\x06status\x18\n \x01(\x0e\x32\x36.tecton_proto.modelartifactservice.ModelArtifactStatusR\x06status\x12\x39\n\ncreated_at\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x39\n\nupdated_at\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tupdatedAt\x12&\n\x0fmodel_file_path\x18\r \x01(\tR\rmodelFilePath\x12\x33\n\x16model_config_file_path\x18\x11 \x01(\tR\x13modelConfigFilePath\x12%\n\x0e\x61rtifact_files\x18\x12 \x03(\tR\rartifactFiles\x12\"\n\x0c\x65nvironments\x18\x0e \x03(\tR\x0c\x65nvironments\x12;\n\ncreated_by\x18\x0f \x01(\x0b\x32\x1c.tecton_proto.auth.PrincipalR\tcreatedBy\x12S\n\x14\x63reated_by_principal\x18\x10 \x01(\x0b\x32!.tecton_proto.auth.PrincipalBasicR\x12\x63reatedByPrincipal\x1a=\n\x0f\x46ileHashesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xa2\x01\n\tModelInfo\x12\'\n\x02id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12(\n\x10model_public_uri\x18\x03 \x01(\tR\x0emodelPublicUri\x12.\n\x13metadata_public_uri\x18\x04 \x01(\tR\x11metadataPublicUri*t\n\x13ModelArtifactStatus\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12\x10\n\x0cPENDING_FILE\x10\x01\x12\x10\n\x0cPENDING_SCAN\x10\x02\x12\t\n\x05READY\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x12\x0b\n\x07\x44\x45LETED\x10\x05*[\n\tModelType\x12\x1a\n\x16MODEL_TYPE_UNSPECIFIED\x10\x00\x12\x19\n\x15TECTON_TEXT_EMBEDDING\x10\x01\x12\x0b\n\x07PYTORCH\x10\x02\x12\n\n\x06PYTHON\x10\x03\x42u\n\x1f\x63om.tecton.modelartifactserviceB\x18ModelArtifactServiceDataP\x01Z6github.com/tecton-ai/tecton_proto/modelartifactservice')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.modelartifactservice.model_artifact_data__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037com.tecton.modelartifactserviceB\030ModelArtifactServiceDataP\001Z6github.com/tecton-ai/tecton_proto/modelartifactservice'
  _MODELARTIFACTINFO_FILEHASHESENTRY._options = None
  _MODELARTIFACTINFO_FILEHASHESENTRY._serialized_options = b'8\001'
  _MODELARTIFACTINFO_TAGSENTRY._options = None
  _MODELARTIFACTINFO_TAGSENTRY._serialized_options = b'8\001'
  _MODELARTIFACTSTATUS._serialized_start=1594
  _MODELARTIFACTSTATUS._serialized_end=1710
  _MODELTYPE._serialized_start=1712
  _MODELTYPE._serialized_end=1803
  _MODELARTIFACTINFO._serialized_start=263
  _MODELARTIFACTINFO._serialized_end=1427
  _MODELARTIFACTINFO_FILEHASHESENTRY._serialized_start=1309
  _MODELARTIFACTINFO_FILEHASHESENTRY._serialized_end=1370
  _MODELARTIFACTINFO_TAGSENTRY._serialized_start=1372
  _MODELARTIFACTINFO_TAGSENTRY._serialized_end=1427
  _MODELINFO._serialized_start=1430
  _MODELINFO._serialized_end=1592
# @@protoc_insertion_point(module_scope)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.auth import service__client_pb2 as _service__client_pb2
from tecton_proto.common import server_group_status__client_pb2 as _server_group_status__client_pb2
from tecton_proto.common import server_group_type__client_pb2 as _server_group_type__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetServerGroupRequest(_message.Message):
    __slots__ = ["server_group_name", "workspace"]
    SERVER_GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    server_group_name: str
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., server_group_name: _Optional[str] = ...) -> None: ...

class GetServerGroupResponse(_message.Message):
    __slots__ = ["server_group"]
    SERVER_GROUP_FIELD_NUMBER: _ClassVar[int]
    server_group: ServerGroupInfo
    def __init__(self, server_group: _Optional[_Union[ServerGroupInfo, _Mapping]] = ...) -> None: ...

class ListServerGroupsRequest(_message.Message):
    __slots__ = ["type", "workspace"]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    type: _server_group_type__client_pb2.ServerGroupType
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., type: _Optional[_Union[_server_group_type__client_pb2.ServerGroupType, str]] = ...) -> None: ...

class ListServerGroupsResponse(_message.Message):
    __slots__ = ["server_groups"]
    SERVER_GROUPS_FIELD_NUMBER: _ClassVar[int]
    server_groups: _containers.RepeatedCompositeFieldContainer[ServerGroupInfo]
    def __init__(self, server_groups: _Optional[_Iterable[_Union[ServerGroupInfo, _Mapping]]] = ...) -> None: ...

class ServerGroupInfo(_message.Message):
    __slots__ = ["created_at", "current_config", "description", "desired_config", "environment", "last_modified_by", "name", "owner", "server_group_id", "status", "status_details", "tags", "type"]
    class TagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CURRENT_CONFIG_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    DESIRED_CONFIG_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    LAST_MODIFIED_BY_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    OWNER_FIELD_NUMBER: _ClassVar[int]
    SERVER_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    STATUS_DETAILS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    created_at: _timestamp_pb2.Timestamp
    current_config: ServerGroupScalingConfig
    description: str
    desired_config: ServerGroupScalingConfig
    environment: str
    last_modified_by: str
    name: str
    owner: str
    server_group_id: str
    status: _server_group_status__client_pb2.ServerGroupStatus
    status_details: str
    tags: _containers.ScalarMap[str, str]
    type: _server_group_type__client_pb2.ServerGroupType
    def __init__(self, server_group_id: _Optional[str] = ..., name: _Optional[str] = ..., type: _Optional[_Union[_server_group_type__client_pb2.ServerGroupType, str]] = ..., description: _Optional[str] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., owner: _Optional[str] = ..., last_modified_by: _Optional[str] = ..., tags: _Optional[_Mapping[str, str]] = ..., desired_config: _Optional[_Union[ServerGroupScalingConfig, _Mapping]] = ..., current_config: _Optional[_Union[ServerGroupScalingConfig, _Mapping]] = ..., status: _Optional[_Union[_server_group_status__client_pb2.ServerGroupStatus, str]] = ..., status_details: _Optional[str] = ..., environment: _Optional[str] = ...) -> None: ...

class ServerGroupScalingConfig(_message.Message):
    __slots__ = ["autoscaling_enabled", "desired_nodes", "last_updated_at", "max_nodes", "min_nodes", "workspace_state_id"]
    AUTOSCALING_ENABLED_FIELD_NUMBER: _ClassVar[int]
    DESIRED_NODES_FIELD_NUMBER: _ClassVar[int]
    LAST_UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    MAX_NODES_FIELD_NUMBER: _ClassVar[int]
    MIN_NODES_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_STATE_ID_FIELD_NUMBER: _ClassVar[int]
    autoscaling_enabled: bool
    desired_nodes: int
    last_updated_at: _timestamp_pb2.Timestamp
    max_nodes: int
    min_nodes: int
    workspace_state_id: str
    def __init__(self, min_nodes: _Optional[int] = ..., max_nodes: _Optional[int] = ..., desired_nodes: _Optional[int] = ..., autoscaling_enabled: bool = ..., last_updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., workspace_state_id: _Optional[str] = ...) -> None: ...



import sys
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

from tecton_core import materialization_context
from tecton_core.errors import TectonValidationError
from tecton_proto.args.user_defined_function__client_pb2 import UserDefinedFunction


# TODO(deprecated_after=0.5): handle backward-compatibility for builtin transformations that did not use tecton.materialization_context
# but instead directly accessed tecton_spark.materialization_context
sys.modules["tecton_spark.materialization_context"] = materialization_context


def _from_proto(
    serialized_transform: UserDefinedFunction,
    globals_: Optional[Dict[str, Any]] = None,
    locals_: Optional[Dict[str, Any]] = None,
) -> Callable:
    """
    deserialize into the provided scope, by default we deserialize the functions into their own scopes
    """

    if globals_ is None:
        globals_ = {}

    assert serialized_transform.HasField("body") and serialized_transform.HasField(
        "name"
    ), "Invalid UserDefinedFunction."

    try:
        exec(serialized_transform.body, globals_, locals_)
    except NameError as e:
        msg = "Failed to serialize function. Please note that all imports must be in the body of the function (not top-level) and type annotations cannot require imports. Additionally, be cautious of variables that shadow other variables. See https://docs.tecton.ai/docs/defining-features/feature-views/transformations for more details."
        raise TectonValidationError(
            msg,
            e,
        )

    # Return function pointer
    try:
        fn = eval(serialized_transform.name, globals_, locals_)
        fn._code = serialized_transform.body
        return fn
    except Exception as e:
        msg = "Invalid transform"
        raise ValueError(msg) from e


def _from_proto_pollute_main(serialized_transform: UserDefinedFunction) -> Callable:
    """Deserialize transform using main scope

    This version of function deserialization to __main__.__dict__. This has historically been the behavior of function
    deserialization. Generally this should be avoided since it can cause hard to debug issues, e.g. two helper functions
    of the same name can shadow each other, or allows for users to depend on libraries that they did not import themselves.
    """
    main_scope = __import__("__main__").__dict__
    # deserialize directly to main and modifies that namespace for all future functions :(
    return _from_proto(serialized_transform, globals_=main_scope)


def _from_proto_isolate(serialized_transform: UserDefinedFunction) -> Callable:
    """Deserialize transform with isolated context, but keep access to `spark` and `sc`.

    This version of function deserialization allows user to have access to the `spark` and `sc` variables from __main__
    scope if they exist, but also isolates functions from each other. The "spark" (SparkSession) and "sc" (SparkContext)
    variables are provided by default in shell environments and DataBricks Notebooks, so users may be relying on them.
    This version ensures these variables are still available without deserializing the function _into_ the main context.
    """
    deserialization_globals = {}
    main_scope = __import__("__main__").__dict__
    for var in ["spark", "sc"]:
        if var in main_scope:
            deserialization_globals[var] = main_scope.get(var)

    return _from_proto(serialized_transform, globals_=deserialization_globals)


def _from_proto_empty_globals(serialized_transform: UserDefinedFunction) -> Callable:
    """Deserialize transform with an empty globals

    Do not deserialize the function into __main__ context (good), do not import other random additional things into
    but also means users may lose access to any variables from __main__ they might have been relying on.
    """
    return _from_proto(serialized_transform, globals_={})


def from_proto(serialized_transform: UserDefinedFunction, include_main_variable_in_scope: bool = False) -> Callable:
    """Takes in a Serialized function and converts it into a python function.

    :param serialized_transform: A UserDefinedFunction proto object with the serialized function and additional config
    :param include_main_variable_in_scope: Whether to deserialize with globals 'empty', or ensure that certain variables
        such as 'spark' or 'sc' are passed in from the global context. This differs depending on if you are using
        materialization, local, etc.
    :return: The deserialized python function
    """
    flag_value = serialized_transform.isolate_function_deserialization
    if include_main_variable_in_scope and flag_value:
        deserialized_function = _from_proto_isolate(serialized_transform)
    elif include_main_variable_in_scope:
        deserialized_function = _from_proto_pollute_main(serialized_transform)
    else:
        deserialized_function = _from_proto_empty_globals(serialized_transform)

    return deserialized_function



import inspect
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import attrs
from typeguard import typechecked

from tecton._internals import errors
from tecton._internals import sdk_decorators
from tecton._internals import validations_api
from tecton._internals.repo import function_serialization
from tecton.framework import base_tecton_object
from tecton.framework.configs import Secret
from tecton.framework.configs import convert_secret_to_sanitized_reference
from tecton_core import conf
from tecton_core import id_helper
from tecton_core import specs
from tecton_core.repo_file_handler import construct_fco_source_info
from tecton_proto.args import basic_info__client_pb2 as basic_info_pb2
from tecton_proto.args import fco_args__client_pb2 as fco_args_pb2
from tecton_proto.args import resource_provider__client_pb2 as resource_provider__args_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


RESOURCE_CONTEXT_PARAM_NAME = "context"


@attrs.define(eq=False)
class ResourceProvider(base_tecton_object.BaseTectonObject):
    """
    A `ResourceProvider` provides a stateful resource in Tecton, It is initialized in advance so it can be reused among certain computations such as materialization job.

    ```python
    from tecton import resource_provider, Secret

    @resource_provider(
        tags = {"environment":"staging"},
        owner = "tom@tecton.ai",
        secrets = {"api_key": Secret(scope="scope", key="api_key")}
    )
    def resource(context):
        return APIClient(api_key = context.secrets["api_key"])
    ```

    """

    # A Tecton "args" proto. Only set if this object was defined locally, i.e. this object was not applied and fetched
    # from the Tecton backend.
    _args: Optional[resource_provider__args_pb2.ResourceProviderArgs] = attrs.field(
        repr=False, on_setattr=attrs.setters.frozen
    )

    _spec: Optional[specs.ResourceProviderSpec] = attrs.field(repr=False)

    def __init__(
        self,
        *,
        name: str,
        function: Callable,
        description: Optional[str] = None,
        owner: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        secrets: Optional[Dict[str, Union[Secret, str]]] = None,
    ):
        param_names = [param.name for param in inspect.signature(function).parameters.values()]
        if param_names and param_names != [RESOURCE_CONTEXT_PARAM_NAME]:
            raise errors.INVALID_INPUTS_RESOURCE_PROVIDER(name)

        serialized_function = None
        if function_serialization.should_serialize_function(function):
            serialized_function = function_serialization.to_proto(function)
        secret_references = {}
        if secrets:
            for secret_name, secret in secrets.items():
                secret_references[secret_name] = convert_secret_to_sanitized_reference(secret)

        args = resource_provider__args_pb2.ResourceProviderArgs(
            resource_provider_id=id_helper.IdHelper.generate_id(),
            info=basic_info_pb2.BasicInfo(name=name, description=description, tags=tags, owner=owner),
            secrets=secret_references,
            function=serialized_function,
        )
        info = base_tecton_object.TectonObjectInfo.from_args_proto(args.info, args.resource_provider_id)
        source_info = construct_fco_source_info(args.resource_provider_id)
        self.__attrs_init__(
            info=info,
            args=args,
            source_info=source_info,
            spec=None,
        )
        if not conf.get_bool("TECTON_SKIP_OBJECT_VALIDATION"):
            self._validate()
        self._spec = specs.ResourceProviderSpec.from_args_proto(self._args, function)
        base_tecton_object._register_local_object(self)

    def _validate(self) -> None:
        validations_api.run_backend_validation_and_assert_valid(
            self,
            validator_pb2.ValidationRequest(validation_args=[self._build_fco_validation_args()]),
        )

    @classmethod
    @typechecked
    def _from_spec(cls, spec: specs.ResourceProviderSpec) -> "ResourceProvider":
        """Create an ResourceProvider directly from a spec. Specs are assumed valid and will not be re-validated."""
        info = base_tecton_object.TectonObjectInfo.from_spec(spec)

        obj = cls.__new__(cls)  # Instantiate the object. Does not call init.
        obj.__attrs_init__(info=info, spec=spec, args=None, source_info=None)
        return obj

    @sdk_decorators.assert_local_object
    def _build_and_resolve_args(self, objects) -> fco_args_pb2.FcoArgs:
        return fco_args_pb2.FcoArgs(resource_provider=self._args)

    def _build_fco_validation_args(self) -> validator_pb2.FcoValidationArgs:
        if self.info._is_local_object:
            return validator_pb2.FcoValidationArgs(
                resource_provider=validator_pb2.ResourceProviderArgs(
                    args=self._args,
                )
            )
        else:
            return self._spec.validation_args


def resource_provider(
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    owner: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    secrets: Optional[Dict[str, Union[Secret, str]]] = None,
):
    """
    Declare a ResourceProvider

    :param name: Unique name for resource provider, Defaults to function name.
    :param description: Description of Resource Provider
    :param owner: Owner of resource provider (typically email)
    :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
    :param secrets: A dictionary of Secret references that will be resolved and provided to the resource provider at runtime of the Feature View that this resource is used in. During local development and testing, strings may be used instead Secret references.

    :return: An object of type `ResourceProvider`
    """

    def decorator(function):
        return ResourceProvider(
            name=name or function.__name__,
            description=description,
            owner=owner,
            tags=tags,
            secrets=secrets,
            function=function,
        )

    return decorator



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.args import feature_view__client_pb2 as _feature_view__client_pb2
from tecton_proto.auth import service__client_pb2 as _service__client_pb2
from tecton_proto.common import compute_mode__client_pb2 as _compute_mode__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import schema__client_pb2 as _schema__client_pb2
from tecton_proto.materialization import spark_cluster__client_pb2 as _spark_cluster__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
TEST_ONLY_MATERIALIZATION_JOB_TYPE_BATCH: TestOnlyMaterializationJobType
TEST_ONLY_MATERIALIZATION_JOB_TYPE_DATASET_GENERATION: TestOnlyMaterializationJobType
TEST_ONLY_MATERIALIZATION_JOB_TYPE_INGEST: TestOnlyMaterializationJobType
TEST_ONLY_MATERIALIZATION_JOB_TYPE_MAINTENANCE: TestOnlyMaterializationJobType
TEST_ONLY_MATERIALIZATION_JOB_TYPE_STREAM: TestOnlyMaterializationJobType
TEST_ONLY_MATERIALIZATION_JOB_TYPE_UNSPECIFIED: TestOnlyMaterializationJobType

class CancelDatasetJobRequest(_message.Message):
    __slots__ = ["job_id", "saved_feature_data_frame", "workspace"]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    SAVED_FEATURE_DATA_FRAME_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    saved_feature_data_frame: str
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., saved_feature_data_frame: _Optional[str] = ..., job_id: _Optional[str] = ...) -> None: ...

class CancelDatasetJobResponse(_message.Message):
    __slots__ = ["job"]
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: MaterializationJob
    def __init__(self, job: _Optional[_Union[MaterializationJob, _Mapping]] = ...) -> None: ...

class CancelJobRequest(_message.Message):
    __slots__ = ["feature_service", "feature_view", "job_id", "workspace"]
    FEATURE_SERVICE_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: _ClassVar[int]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    feature_service: str
    feature_view: str
    job_id: str
    workspace: str
    def __init__(self, job_id: _Optional[str] = ..., workspace: _Optional[str] = ..., feature_view: _Optional[str] = ..., feature_service: _Optional[str] = ...) -> None: ...

class CancelJobResponse(_message.Message):
    __slots__ = ["job"]
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: MaterializationJob
    def __init__(self, job: _Optional[_Union[MaterializationJob, _Mapping]] = ...) -> None: ...

class CompleteDataframeUploadRequest(_message.Message):
    __slots__ = ["key", "part_etags", "upload_id", "workspace"]
    class PartEtagsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: int
        value: str
        def __init__(self, key: _Optional[int] = ..., value: _Optional[str] = ...) -> None: ...
    KEY_FIELD_NUMBER: _ClassVar[int]
    PART_ETAGS_FIELD_NUMBER: _ClassVar[int]
    UPLOAD_ID_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    key: str
    part_etags: _containers.ScalarMap[int, str]
    upload_id: str
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., key: _Optional[str] = ..., upload_id: _Optional[str] = ..., part_etags: _Optional[_Mapping[int, str]] = ...) -> None: ...

class CompleteDataframeUploadResponse(_message.Message):
    __slots__ = ["key"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    key: str
    def __init__(self, key: _Optional[str] = ...) -> None: ...

class GetDataframeInfoRequest(_message.Message):
    __slots__ = ["feature_service", "feature_view", "task_type", "workspace"]
    FEATURE_SERVICE_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: _ClassVar[int]
    TASK_TYPE_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    feature_service: str
    feature_view: str
    task_type: _spark_cluster__client_pb2.TaskType
    workspace: str
    def __init__(self, feature_view: _Optional[str] = ..., feature_service: _Optional[str] = ..., workspace: _Optional[str] = ..., task_type: _Optional[_Union[_spark_cluster__client_pb2.TaskType, str]] = ...) -> None: ...

class GetDataframeInfoResponse(_message.Message):
    __slots__ = ["df_path", "signed_url_for_df_upload"]
    DF_PATH_FIELD_NUMBER: _ClassVar[int]
    SIGNED_URL_FOR_DF_UPLOAD_FIELD_NUMBER: _ClassVar[int]
    df_path: str
    signed_url_for_df_upload: str
    def __init__(self, df_path: _Optional[str] = ..., signed_url_for_df_upload: _Optional[str] = ...) -> None: ...

class GetDataframeUploadUrlRequest(_message.Message):
    __slots__ = ["feature_service", "feature_view", "task_type", "workspace"]
    FEATURE_SERVICE_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: _ClassVar[int]
    TASK_TYPE_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    feature_service: str
    feature_view: str
    task_type: _spark_cluster__client_pb2.TaskType
    workspace: str
    def __init__(self, feature_view: _Optional[str] = ..., feature_service: _Optional[str] = ..., workspace: _Optional[str] = ..., task_type: _Optional[_Union[_spark_cluster__client_pb2.TaskType, str]] = ...) -> None: ...

class GetDataframeUploadUrlResponse(_message.Message):
    __slots__ = ["key", "upload_id"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    UPLOAD_ID_FIELD_NUMBER: _ClassVar[int]
    key: str
    upload_id: str
    def __init__(self, key: _Optional[str] = ..., upload_id: _Optional[str] = ...) -> None: ...

class GetDatasetJobRequest(_message.Message):
    __slots__ = ["job_id", "saved_feature_data_frame", "workspace"]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    SAVED_FEATURE_DATA_FRAME_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    saved_feature_data_frame: str
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., saved_feature_data_frame: _Optional[str] = ..., job_id: _Optional[str] = ...) -> None: ...

class GetDatasetJobResponse(_message.Message):
    __slots__ = ["job"]
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: MaterializationJob
    def __init__(self, job: _Optional[_Union[MaterializationJob, _Mapping]] = ...) -> None: ...

class GetJobRequest(_message.Message):
    __slots__ = ["feature_service", "feature_view", "job_id", "workspace"]
    FEATURE_SERVICE_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: _ClassVar[int]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    feature_service: str
    feature_view: str
    job_id: str
    workspace: str
    def __init__(self, job_id: _Optional[str] = ..., workspace: _Optional[str] = ..., feature_view: _Optional[str] = ..., feature_service: _Optional[str] = ...) -> None: ...

class GetJobResponse(_message.Message):
    __slots__ = ["job"]
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: MaterializationJob
    def __init__(self, job: _Optional[_Union[MaterializationJob, _Mapping]] = ...) -> None: ...

class GetLatestReadyTimeRequest(_message.Message):
    __slots__ = ["feature_service", "feature_view", "workspace"]
    FEATURE_SERVICE_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    feature_service: str
    feature_view: str
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., feature_view: _Optional[str] = ..., feature_service: _Optional[str] = ...) -> None: ...

class GetLatestReadyTimeResponse(_message.Message):
    __slots__ = ["offline_latest_ready_time", "online_latest_ready_time"]
    OFFLINE_LATEST_READY_TIME_FIELD_NUMBER: _ClassVar[int]
    ONLINE_LATEST_READY_TIME_FIELD_NUMBER: _ClassVar[int]
    offline_latest_ready_time: _timestamp_pb2.Timestamp
    online_latest_ready_time: _timestamp_pb2.Timestamp
    def __init__(self, online_latest_ready_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., offline_latest_ready_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class IngestDataframeFromS3Request(_message.Message):
    __slots__ = ["df_path", "feature_view", "use_tecton_managed_retries", "workspace"]
    DF_PATH_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: _ClassVar[int]
    USE_TECTON_MANAGED_RETRIES_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    df_path: str
    feature_view: str
    use_tecton_managed_retries: bool
    workspace: str
    def __init__(self, feature_view: _Optional[str] = ..., df_path: _Optional[str] = ..., workspace: _Optional[str] = ..., use_tecton_managed_retries: bool = ...) -> None: ...

class IngestDataframeFromS3Response(_message.Message):
    __slots__ = ["job"]
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: MaterializationJob
    def __init__(self, job: _Optional[_Union[MaterializationJob, _Mapping]] = ...) -> None: ...

class JobAttempt(_message.Message):
    __slots__ = ["created_at", "id", "run_url", "state", "updated_at"]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    RUN_URL_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    created_at: _timestamp_pb2.Timestamp
    id: str
    run_url: str
    state: str
    updated_at: _timestamp_pb2.Timestamp
    def __init__(self, id: _Optional[str] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., state: _Optional[str] = ..., run_url: _Optional[str] = ...) -> None: ...

class ListJobsRequest(_message.Message):
    __slots__ = ["feature_service", "feature_view", "workspace"]
    FEATURE_SERVICE_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    feature_service: str
    feature_view: str
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., feature_view: _Optional[str] = ..., feature_service: _Optional[str] = ...) -> None: ...

class ListJobsResponse(_message.Message):
    __slots__ = ["jobs"]
    JOBS_FIELD_NUMBER: _ClassVar[int]
    jobs: _containers.RepeatedCompositeFieldContainer[MaterializationJob]
    def __init__(self, jobs: _Optional[_Iterable[_Union[MaterializationJob, _Mapping]]] = ...) -> None: ...

class MaterializationJob(_message.Message):
    __slots__ = ["attempts", "created_at", "end_time", "feature_service", "feature_view", "id", "ingest_path", "job_type", "next_attempt_at", "offline", "online", "saved_feature_data_frame", "start_time", "state", "updated_at", "workspace"]
    ATTEMPTS_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    FEATURE_SERVICE_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    INGEST_PATH_FIELD_NUMBER: _ClassVar[int]
    JOB_TYPE_FIELD_NUMBER: _ClassVar[int]
    NEXT_ATTEMPT_AT_FIELD_NUMBER: _ClassVar[int]
    OFFLINE_FIELD_NUMBER: _ClassVar[int]
    ONLINE_FIELD_NUMBER: _ClassVar[int]
    SAVED_FEATURE_DATA_FRAME_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    attempts: _containers.RepeatedCompositeFieldContainer[JobAttempt]
    created_at: _timestamp_pb2.Timestamp
    end_time: _timestamp_pb2.Timestamp
    feature_service: str
    feature_view: str
    id: str
    ingest_path: str
    job_type: str
    next_attempt_at: _timestamp_pb2.Timestamp
    offline: bool
    online: bool
    saved_feature_data_frame: str
    start_time: _timestamp_pb2.Timestamp
    state: str
    updated_at: _timestamp_pb2.Timestamp
    workspace: str
    def __init__(self, id: _Optional[str] = ..., workspace: _Optional[str] = ..., feature_view: _Optional[str] = ..., feature_service: _Optional[str] = ..., saved_feature_data_frame: _Optional[str] = ..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., state: _Optional[str] = ..., attempts: _Optional[_Iterable[_Union[JobAttempt, _Mapping]]] = ..., next_attempt_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., online: bool = ..., offline: bool = ..., job_type: _Optional[str] = ..., ingest_path: _Optional[str] = ...) -> None: ...

class MaterializationJobRequest(_message.Message):
    __slots__ = ["end_time", "feature_view", "offline", "online", "overwrite", "start_time", "use_tecton_managed_retries", "workspace"]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_FIELD_NUMBER: _ClassVar[int]
    OFFLINE_FIELD_NUMBER: _ClassVar[int]
    ONLINE_FIELD_NUMBER: _ClassVar[int]
    OVERWRITE_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    USE_TECTON_MANAGED_RETRIES_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    end_time: _timestamp_pb2.Timestamp
    feature_view: str
    offline: bool
    online: bool
    overwrite: bool
    start_time: _timestamp_pb2.Timestamp
    use_tecton_managed_retries: bool
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., feature_view: _Optional[str] = ..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., online: bool = ..., offline: bool = ..., use_tecton_managed_retries: bool = ..., overwrite: bool = ...) -> None: ...

class MaterializationJobResponse(_message.Message):
    __slots__ = ["job"]
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: MaterializationJob
    def __init__(self, job: _Optional[_Union[MaterializationJob, _Mapping]] = ...) -> None: ...

class StartDatasetJobRequest(_message.Message):
    __slots__ = ["cluster_config", "compute_mode", "dataset_name", "datetime_range", "environment", "expected_schema", "extra_config", "feature_service_id", "feature_view_id", "from_source", "spine", "tecton_runtime", "workspace"]
    class DateTimeRangeInput(_message.Message):
        __slots__ = ["end", "entities_path", "max_lookback", "start"]
        END_FIELD_NUMBER: _ClassVar[int]
        ENTITIES_PATH_FIELD_NUMBER: _ClassVar[int]
        MAX_LOOKBACK_FIELD_NUMBER: _ClassVar[int]
        START_FIELD_NUMBER: _ClassVar[int]
        end: _timestamp_pb2.Timestamp
        entities_path: str
        max_lookback: _timestamp_pb2.Timestamp
        start: _timestamp_pb2.Timestamp
        def __init__(self, start: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., end: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., max_lookback: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., entities_path: _Optional[str] = ...) -> None: ...
    class ExtraConfigEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class SpineInput(_message.Message):
        __slots__ = ["column_names", "path", "timestamp_key"]
        COLUMN_NAMES_FIELD_NUMBER: _ClassVar[int]
        PATH_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMP_KEY_FIELD_NUMBER: _ClassVar[int]
        column_names: _containers.RepeatedScalarFieldContainer[str]
        path: str
        timestamp_key: str
        def __init__(self, path: _Optional[str] = ..., timestamp_key: _Optional[str] = ..., column_names: _Optional[_Iterable[str]] = ...) -> None: ...
    CLUSTER_CONFIG_FIELD_NUMBER: _ClassVar[int]
    COMPUTE_MODE_FIELD_NUMBER: _ClassVar[int]
    DATASET_NAME_FIELD_NUMBER: _ClassVar[int]
    DATETIME_RANGE_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    EXPECTED_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    EXTRA_CONFIG_FIELD_NUMBER: _ClassVar[int]
    FEATURE_SERVICE_ID_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_ID_FIELD_NUMBER: _ClassVar[int]
    FROM_SOURCE_FIELD_NUMBER: _ClassVar[int]
    SPINE_FIELD_NUMBER: _ClassVar[int]
    TECTON_RUNTIME_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    cluster_config: _feature_view__client_pb2.ClusterConfig
    compute_mode: _compute_mode__client_pb2.BatchComputeMode
    dataset_name: str
    datetime_range: StartDatasetJobRequest.DateTimeRangeInput
    environment: str
    expected_schema: _schema__client_pb2.Schema
    extra_config: _containers.ScalarMap[str, str]
    feature_service_id: _id__client_pb2.Id
    feature_view_id: _id__client_pb2.Id
    from_source: bool
    spine: StartDatasetJobRequest.SpineInput
    tecton_runtime: str
    workspace: str
    def __init__(self, compute_mode: _Optional[_Union[_compute_mode__client_pb2.BatchComputeMode, str]] = ..., from_source: bool = ..., workspace: _Optional[str] = ..., feature_service_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., feature_view_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., spine: _Optional[_Union[StartDatasetJobRequest.SpineInput, _Mapping]] = ..., datetime_range: _Optional[_Union[StartDatasetJobRequest.DateTimeRangeInput, _Mapping]] = ..., dataset_name: _Optional[str] = ..., cluster_config: _Optional[_Union[_feature_view__client_pb2.ClusterConfig, _Mapping]] = ..., tecton_runtime: _Optional[str] = ..., environment: _Optional[str] = ..., extra_config: _Optional[_Mapping[str, str]] = ..., expected_schema: _Optional[_Union[_schema__client_pb2.Schema, _Mapping]] = ...) -> None: ...

class StartDatasetJobResponse(_message.Message):
    __slots__ = ["job"]
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: MaterializationJob
    def __init__(self, job: _Optional[_Union[MaterializationJob, _Mapping]] = ...) -> None: ...

class TestOnlyCompleteOnlineTableRequest(_message.Message):
    __slots__ = ["feature_view_name", "workspace"]
    FEATURE_VIEW_NAME_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    feature_view_name: str
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., feature_view_name: _Optional[str] = ...) -> None: ...

class TestOnlyCompleteOnlineTableResponse(_message.Message):
    __slots__ = ["table_name"]
    TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    table_name: str
    def __init__(self, table_name: _Optional[str] = ...) -> None: ...

class TestOnlyGetDatasetGenerationTaskParamsRequest(_message.Message):
    __slots__ = ["start_dataset_job_request"]
    START_DATASET_JOB_REQUEST_FIELD_NUMBER: _ClassVar[int]
    start_dataset_job_request: StartDatasetJobRequest
    def __init__(self, start_dataset_job_request: _Optional[_Union[StartDatasetJobRequest, _Mapping]] = ...) -> None: ...

class TestOnlyGetMaterializationTaskParamsRequest(_message.Message):
    __slots__ = ["df_path", "disable_offline", "disable_online", "feature_view_name", "job_end_time", "job_start_time", "job_type", "workspace"]
    DF_PATH_FIELD_NUMBER: _ClassVar[int]
    DISABLE_OFFLINE_FIELD_NUMBER: _ClassVar[int]
    DISABLE_ONLINE_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VIEW_NAME_FIELD_NUMBER: _ClassVar[int]
    JOB_END_TIME_FIELD_NUMBER: _ClassVar[int]
    JOB_START_TIME_FIELD_NUMBER: _ClassVar[int]
    JOB_TYPE_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    df_path: str
    disable_offline: bool
    disable_online: bool
    feature_view_name: str
    job_end_time: _timestamp_pb2.Timestamp
    job_start_time: _timestamp_pb2.Timestamp
    job_type: TestOnlyMaterializationJobType
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., feature_view_name: _Optional[str] = ..., job_type: _Optional[_Union[TestOnlyMaterializationJobType, str]] = ..., disable_offline: bool = ..., disable_online: bool = ..., job_start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., job_end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., df_path: _Optional[str] = ...) -> None: ...

class TestOnlyGetMaterializationTaskParamsResponse(_message.Message):
    __slots__ = ["encoded_materialization_params"]
    ENCODED_MATERIALIZATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    encoded_materialization_params: str
    def __init__(self, encoded_materialization_params: _Optional[str] = ...) -> None: ...

class TestOnlyOnlineTableNameRequest(_message.Message):
    __slots__ = ["feature_view_name", "watermark", "workspace"]
    FEATURE_VIEW_NAME_FIELD_NUMBER: _ClassVar[int]
    WATERMARK_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    feature_view_name: str
    watermark: _timestamp_pb2.Timestamp
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., feature_view_name: _Optional[str] = ..., watermark: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class TestOnlyOnlineTableNameResponse(_message.Message):
    __slots__ = ["table_name"]
    TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    table_name: str
    def __init__(self, table_name: _Optional[str] = ...) -> None: ...

class TestOnlyWriteFeatureServerConfigRequest(_message.Message):
    __slots__ = ["absolute_filepath"]
    ABSOLUTE_FILEPATH_FIELD_NUMBER: _ClassVar[int]
    absolute_filepath: str
    def __init__(self, absolute_filepath: _Optional[str] = ...) -> None: ...

class TestOnlyWriteFeatureServerConfigResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class UploadDataframePartRequest(_message.Message):
    __slots__ = ["key", "parent_upload_id", "part_number", "workspace"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    PARENT_UPLOAD_ID_FIELD_NUMBER: _ClassVar[int]
    PART_NUMBER_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    key: str
    parent_upload_id: str
    part_number: int
    workspace: str
    def __init__(self, workspace: _Optional[str] = ..., key: _Optional[str] = ..., parent_upload_id: _Optional[str] = ..., part_number: _Optional[int] = ...) -> None: ...

class UploadDataframePartResponse(_message.Message):
    __slots__ = ["upload_url"]
    UPLOAD_URL_FIELD_NUMBER: _ClassVar[int]
    upload_url: str
    def __init__(self, upload_url: _Optional[str] = ...) -> None: ...

class TestOnlyMaterializationJobType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class DbfsDeleteRequest(_message.Message):
    __slots__ = ["path", "recursive"]
    PATH_FIELD_NUMBER: _ClassVar[int]
    RECURSIVE_FIELD_NUMBER: _ClassVar[int]
    path: str
    recursive: bool
    def __init__(self, path: _Optional[str] = ..., recursive: bool = ...) -> None: ...

class DbfsGetStatusRequest(_message.Message):
    __slots__ = ["path"]
    PATH_FIELD_NUMBER: _ClassVar[int]
    path: str
    def __init__(self, path: _Optional[str] = ...) -> None: ...

class DbfsReadRequest(_message.Message):
    __slots__ = ["length", "offset", "path"]
    LENGTH_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    length: int
    offset: int
    path: str
    def __init__(self, path: _Optional[str] = ..., offset: _Optional[int] = ..., length: _Optional[int] = ...) -> None: ...

class DbfsReadResponse(_message.Message):
    __slots__ = ["bytes_read", "data"]
    BYTES_READ_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    bytes_read: int
    data: bytes
    def __init__(self, bytes_read: _Optional[int] = ..., data: _Optional[bytes] = ...) -> None: ...



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class StageLocation(_message.Message):
    __slots__ = ["namespace", "path", "stage_name"]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    STAGE_NAME_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    path: str
    stage_name: str
    def __init__(self, namespace: _Optional[str] = ..., stage_name: _Optional[str] = ..., path: _Optional[str] = ...) -> None: ...



from tecton_proto.args import basic_info__client_pb2 as _basic_info__client_pb2
from tecton_proto.args import diff_options__client_pb2 as _diff_options__client_pb2
from tecton_proto.args import user_defined_function__client_pb2 as _user_defined_function__client_pb2
from tecton_proto.common import framework_version__client_pb2 as _framework_version__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
TRANSFORMATION_MODE_BIGQUERY_SQL: TransformationMode
TRANSFORMATION_MODE_NO_TRANSFORMATION: TransformationMode
TRANSFORMATION_MODE_PANDAS: TransformationMode
TRANSFORMATION_MODE_PYSPARK: TransformationMode
TRANSFORMATION_MODE_PYTHON: TransformationMode
TRANSFORMATION_MODE_SNOWFLAKE_SQL: TransformationMode
TRANSFORMATION_MODE_SNOWPARK: TransformationMode
TRANSFORMATION_MODE_SPARK_SQL: TransformationMode
TRANSFORMATION_MODE_UNSPECIFIED: TransformationMode

class TransformationArgs(_message.Message):
    __slots__ = ["docstring", "info", "is_builtin", "options", "prevent_destroy", "transformation_id", "transformation_mode", "user_function", "version"]
    class OptionsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    DOCSTRING_FIELD_NUMBER: _ClassVar[int]
    INFO_FIELD_NUMBER: _ClassVar[int]
    IS_BUILTIN_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    PREVENT_DESTROY_FIELD_NUMBER: _ClassVar[int]
    TRANSFORMATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRANSFORMATION_MODE_FIELD_NUMBER: _ClassVar[int]
    USER_FUNCTION_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    docstring: str
    info: _basic_info__client_pb2.BasicInfo
    is_builtin: bool
    options: _containers.ScalarMap[str, str]
    prevent_destroy: bool
    transformation_id: _id__client_pb2.Id
    transformation_mode: TransformationMode
    user_function: _user_defined_function__client_pb2.UserDefinedFunction
    version: _framework_version__client_pb2.FrameworkVersion
    def __init__(self, transformation_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., info: _Optional[_Union[_basic_info__client_pb2.BasicInfo, _Mapping]] = ..., version: _Optional[_Union[_framework_version__client_pb2.FrameworkVersion, str]] = ..., prevent_destroy: bool = ..., options: _Optional[_Mapping[str, str]] = ..., transformation_mode: _Optional[_Union[TransformationMode, str]] = ..., user_function: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ..., docstring: _Optional[str] = ..., is_builtin: bool = ...) -> None: ...

class TransformationMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import math
from dataclasses import dataclass
from pathlib import Path
from typing import List


# boto3 defaults to 8MB for multi-part uploads using upload_file.
DEFAULT_UPLOAD_PART_SIZE_MB = 16

# 5 was arbitrarily selected. We want to be conservative as this will run in customer's environments
DEFAULT_MAX_WORKERS_THREADS = 5

MEGABYTE = 1024 * 1024
GIGABYTE = 1024 * MEGABYTE


@dataclass
class UploadPart:
    """
    Represents an individual part of a file that needs to be uploaded in chunks or parts.
    :param part_number (int): The 1-indexed number of the part to be uploaded.
    :param offset (int): The starting byte offset of this part in the file.
    :param part_size (int): The size of this part in bytes.
    """

    part_number: int
    offset: int
    part_size: int


def _calculate_part_count(file_size: int, part_size_mb: int) -> int:
    """Calculate the number of parts the file will be divided into for uploading.

    Args:
        file_size (int): The size of the file
        part_size_mb (int): The size of each part in megabytes.

    Returns:
        int: The total number of parts.
    """
    chunk_size = part_size_mb * 1024 * 1024
    return int(math.ceil(file_size / chunk_size))


def get_upload_parts(file_size: int) -> List[UploadPart]:
    """
    Calculate UploadPart for each part of a file to be uploaded, given total file size.
    It considers the DEFAULT_UPLOAD_PART_SIZE_MB as the maximum size of each part.

    Args:
        file_size (int): The total size of the file being uploaded in bytes.

    Returns:
        List[UploadPart]: An list of UploadPart representing all parts to be uploaded with its part number,
                    starting offset, and size in bytes.
    """
    total_parts = _calculate_part_count(file_size, DEFAULT_UPLOAD_PART_SIZE_MB)
    chunk_size = DEFAULT_UPLOAD_PART_SIZE_MB * MEGABYTE
    upload_parts = []
    for i in range(1, total_parts + 1):
        offset = chunk_size * (i - 1)
        bytes_remaining = file_size - offset
        # Adjust the size for the last part if the remaining bytes are less than the DEFAULT_UPLOAD_PART_SIZE_MB
        current_chunk_size = chunk_size if bytes_remaining > chunk_size else bytes_remaining
        upload_parts.append(UploadPart(part_number=i, offset=offset, part_size=current_chunk_size))
    return upload_parts


def _get_directory_size(directory: Path) -> int:
    """
    Compute the size of a directory in bytes.

    Args:
        directory (Path): The directory path for which to compute the size.

    Returns:
        int: The size of the directory in bytes.
    """
    return sum(f.stat().st_size for f in directory.rglob("*") if f.is_file())
